#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2008,2010  Gary Burton
# Copyright (C) 2008       Craig J. Anderson
# Copyright (C) 2009       Nick Hall
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Adam Stein <adam@csh.rit.edu>
# Copyright (C) 2011-2013  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Specific option handling for a GUI.
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
import os

# -------------------------------------------------------------------------
#
# gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from ..utils import ProgressMeter
from ..pluginmanager import GuiPluginManager
from .. import widgets
from ..managedwindow import ManagedWindow
from ..dialog import OptionDialog
from ..selectors import SelectorFactory
from gramps.gen.errors import HandleError
from gramps.gen.display.name import displayer as _nd
from gramps.gen.display.place import displayer as _pd
from gramps.gen.filters import GenericFilterFactory, GenericFilter, rules
from gramps.gen.constfunc import get_curr_dir
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# ------------------------------------------------------------------------
#
# Dialog window used to select a surname
#
# ------------------------------------------------------------------------
class LastNameDialog(ManagedWindow):
    """
    A dialog that allows the selection of a surname from the database.
    """

    def __init__(self, database, uistate, track, surnames, skip_list=set()):
        ManagedWindow.__init__(self, uistate, track, self, modal=True)
        self.__dlg = Gtk.Dialog(
            transient_for=uistate.window, destroy_with_parent=True, modal=True
        )
        self.__dlg.add_buttons(
            _("_Cancel"), Gtk.ResponseType.REJECT, _("_OK"), Gtk.ResponseType.ACCEPT
        )
        self.set_window(self.__dlg, None, _("Select surname"))
        self.setup_configs("interface.lastnamedialog", 400, 400)

        # build up a container to display all of the people of interest
        self.__model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_INT)
        self.__tree_view = Gtk.TreeView(model=self.__model)
        col1 = Gtk.TreeViewColumn(_("Surname"), Gtk.CellRendererText(), text=0)
        col2 = Gtk.TreeViewColumn(_("Count"), Gtk.CellRendererText(), text=1)
        col1.set_resizable(True)
        col2.set_resizable(True)
        col1.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col2.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col1.set_sort_column_id(0)
        col2.set_sort_column_id(1)
        self.__tree_view.append_column(col1)
        self.__tree_view.append_column(col2)
        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.add(self.__tree_view)
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_shadow_type(Gtk.ShadowType.OUT)
        self.__dlg.vbox.pack_start(scrolled_window, True, True, 0)
        self.show()

        if len(surnames) == 0:
            # we could use database.get_surname_list(), but if we do that
            # all we get is a list of names without a count...therefore
            # we'll traverse the entire database ourself and build up a
            # list that we can use
            #            for name in database.get_surname_list():
            #                self.__model.append([name, 0])

            # build up the list of surnames, keeping track of the count for each
            # name (this can be a lengthy process, so by passing in the
            # dictionary we can be certain we only do this once)
            progress = ProgressMeter(_("Finding Surnames"), parent=uistate.window)
            progress.set_pass(_("Finding surnames"), database.get_number_of_people())
            for person in database.iter_people():
                progress.step()
                key = person.get_primary_name().get_surname()
                count = 0
                if key in surnames:
                    count = surnames[key]
                surnames[key] = count + 1
            progress.close()

        # insert the names and count into the model
        for key in surnames:
            if key.encode("iso-8859-1", "xmlcharrefreplace") not in skip_list:
                self.__model.append([key, surnames[key]])

        # keep the list sorted starting with the most popular last name
        # (but after sorting the whole list alphabetically first, so
        # that surnames with the same number of people will be alphabetical)
        self.__model.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self.__model.set_sort_column_id(1, Gtk.SortType.DESCENDING)

        # the "OK" button should be enabled/disabled based on the selection of
        # a row
        self.__tree_selection = self.__tree_view.get_selection()
        self.__tree_selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        self.__tree_selection.select_path(0)

    def run(self):
        """
        Display the dialog and return the selected surnames when done.
        """
        response = self.__dlg.run()
        surname_set = set()
        if response == Gtk.ResponseType.ACCEPT:
            (mode, paths) = self.__tree_selection.get_selected_rows()
            for path in paths:
                tree_iter = self.__model.get_iter(path)
                surname = self.__model.get_value(tree_iter, 0)
                surname_set.add(surname)
        if response != Gtk.ResponseType.DELETE_EVENT:
            # ManagedWindow: set the parent dialog to be modal again
            self.close()
        return surname_set

    def build_menu_names(self, obj):
        return (_("Select surname"), None)


# -------------------------------------------------------------------------
#
# GuiStringOption class
#
# -------------------------------------------------------------------------
class GuiStringOption(Gtk.Entry):
    """
    This class displays an option that is a simple one-line string.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.StringOption
        @return: nothing
        """
        Gtk.Entry.__init__(self)
        self.__option = option
        self.set_text(self.__option.get_value())

        # Set up signal handlers when the widget value is changed
        # from user interaction or programmatically.  When handling
        # a specific signal, we need to temporarily block the signal
        # that would call the other signal handler.
        self.changekey = self.connect("changed", self.__text_changed)
        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.conkey = self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        self.set_tooltip_text(self.__option.get_help())

    def __text_changed(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the change of the value made by the user.
        """
        self.__option.disable_signals()
        self.__option.set_value(self.get_text())
        self.__option.enable_signals()

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        self.handler_block(self.changekey)
        self.set_text(self.__option.get_value())
        self.handler_unblock(self.changekey)

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option.disconnect(self.conkey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiColorOption class
#
# -------------------------------------------------------------------------
class GuiColorOption(Gtk.ColorButton):
    """
    This class displays an option that allows the selection of a colour.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        self.__option = option
        Gtk.ColorButton.__init__(self)
        rgba = Gdk.RGBA()
        rgba.parse(self.__option.get_value())
        self.set_rgba(rgba)

        # Set up signal handlers when the widget value is changed
        # from user interaction or programmatically.  When handling
        # a specific signal, we need to temporarily block the signal
        # that would call the other signal handler.
        self.changekey = self.connect("color-set", self.__color_changed)
        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.conkey = self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        self.set_tooltip_text(self.__option.get_help())

    def __color_changed(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the change of color made by the user.
        """
        rgba = self.get_rgba()
        value = "#%02x%02x%02x" % (
            int(rgba.red * 255),
            int(rgba.green * 255),
            int(rgba.blue * 255),
        )

        self.__option.disable_signals()
        self.__option.set_value(value)
        self.__option.enable_signals()

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        self.handler_block(self.changekey)
        rgba = Gdk.RGBA()
        rgba.parse(self.__option.get_value())
        self.set_rgba(rgba)
        self.handler_unblock(self.changekey)

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option.disconnect(self.conkey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiNumberOption class
#
# -------------------------------------------------------------------------
class GuiNumberOption(Gtk.SpinButton):
    """
    This class displays an option that is a simple number with defined maximum
    and minimum values.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        self.__option = option

        decimals = 0
        step = self.__option.get_step()
        adj = Gtk.Adjustment(
            value=1,
            lower=self.__option.get_min(),
            upper=self.__option.get_max(),
            step_increment=step,
        )

        # Calculate the number of decimal places if necessary
        if step < 1:
            import math

            decimals = int(math.log10(step) * -1)

        Gtk.SpinButton.__init__(self, adjustment=adj, climb_rate=1, digits=decimals)
        Gtk.SpinButton.set_numeric(self, True)

        self.set_value(self.__option.get_value())

        # Set up signal handlers when the widget value is changed
        # from user interaction or programmatically.  When handling
        # a specific signal, we need to temporarily block the signal
        # that would call the other signal handler.
        self.changekey = self.connect("value_changed", self.__number_changed)
        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.conkey = self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        self.set_tooltip_text(self.__option.get_help())

    def __number_changed(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the change of the value made by the user.
        """
        vtype = type(self.__option.get_value())

        self.__option.set_value(vtype(self.get_value()))

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        self.handler_block(self.changekey)
        self.set_value(self.__option.get_value())
        self.handler_unblock(self.changekey)

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option.disconnect(self.conkey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiTextOption class
#
# -------------------------------------------------------------------------
class GuiTextOption(Gtk.ScrolledWindow):
    """
    This class displays an option that is a multi-line string.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        self.__option = option
        Gtk.ScrolledWindow.__init__(self)
        self.set_shadow_type(Gtk.ShadowType.IN)
        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.set_vexpand(True)

        # Add a TextView
        value = self.__option.get_value()
        gtext = Gtk.TextView()
        gtext.set_size_request(-1, 70)
        gtext.get_buffer().set_text("\n".join(value))
        gtext.set_editable(1)
        gtext.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.add(gtext)
        self.__buff = gtext.get_buffer()

        # Set up signal handlers when the widget value is changed
        # from user interaction or programmatically.  When handling
        # a specific signal, we need to temporarily block the signal
        # that would call the other signal handler.
        self.bufcon = self.__buff.connect("changed", self.__text_changed)
        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        # Required for tooltip
        gtext.add_events(Gdk.EventMask.ENTER_NOTIFY_MASK)
        gtext.add_events(Gdk.EventMask.LEAVE_NOTIFY_MASK)
        gtext.set_tooltip_text(self.__option.get_help())

    def __text_changed(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the change of the value made by the user.
        """
        text_val = str(
            self.__buff.get_text(
                self.__buff.get_start_iter(), self.__buff.get_end_iter(), False
            )
        )

        self.__option.disable_signals()
        self.__option.set_value(text_val.split("\n"))
        self.__option.enable_signals()

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        self.__buff.handler_block(self.bufcon)

        value = self.__option.get_value()

        # Can only set using a string.  If we have a string value,
        # we'll use that.  If not, we'll assume a list and convert
        # it into a single string by assuming each list element
        # is separated by a newline.
        if isinstance(value, str):
            self.__buff.set_text(value)

            # Need to manually call the other handler so that the option
            # value is changed to be a list.  If left as a string,
            # it would be treated as a list, meaning each character
            # becomes a list element -- not what we want.
            self.__text_changed(None)
        else:
            self.__buff.set_text("\n".join(value))

        self.__buff.handler_unblock(self.bufcon)

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option = None

        self.__buff.disconnect(self.bufcon)
        self.__buff = None


# -------------------------------------------------------------------------
#
# GuiBooleanOption class
#
# -------------------------------------------------------------------------
class GuiBooleanOption(Gtk.CheckButton):
    """
    This class displays an option that is a boolean (True or False).
    """

    def __init__(self, option, dbstate, uistate, track, override):
        self.__option = option
        Gtk.CheckButton.__init__(self)
        self.set_label(self.__option.get_label())
        self.set_active(self.__option.get_value())

        # Set up signal handlers when the widget value is changed
        # from user interaction or programmatically.  When handling
        # a specific signal, we need to temporarily block the signal
        # that would call the other signal handler.
        self.changekey = self.connect("toggled", self.__state_changed)
        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.conkey = self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        self.set_tooltip_text(self.__option.get_help())

    def __state_changed(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the change of the value made by the user.
        """
        self.__option.set_value(self.get_active())

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        self.handler_block(self.changekey)
        self.set_active(self.__option.get_value())
        self.handler_unblock(self.changekey)

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option.disconnect(self.conkey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiEnumeratedListOption class
#
# -------------------------------------------------------------------------
class GuiEnumeratedListOption(Gtk.Box):
    """
    This class displays an option that provides a finite number of values.
    Each possible value is assigned a value and a description.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        Gtk.Box.__init__(self)
        evt_box = Gtk.EventBox()
        self.__option = option
        self.__combo = Gtk.ComboBoxText()
        if len(option.get_items()) > 18:
            self.__combo.set_popup_fixed_width(False)
            self.__combo.set_wrap_width(3)
        evt_box.add(self.__combo)
        self.pack_start(evt_box, True, True, 0)

        self.__update_options()

        # Set up signal handlers when the widget value is changed
        # from user interaction or programmatically.  When handling
        # a specific signal, we need to temporarily block the signal
        # that would call the other signal handler.
        self.changekey = self.__combo.connect("changed", self.__selection_changed)
        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.conkey1 = self.__option.connect("options-changed", self.__update_options)
        self.conkey2 = self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        self.set_tooltip_text(self.__option.get_help())

    def __selection_changed(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the change of the value made by the user.
        """
        index = self.__combo.get_active()
        if index < 0:
            return
        items = self.__option.get_items()
        value, description = items[index]  # IGNORE:W0612 - description is unused

        # Don't disable the __option signals as is normally done for
        # the other widgets or bad things happen (like other needed
        # signals don't fire)

        self.__option.set_value(value)
        self.value_changed()  # Allow overriding so that another class
        # can add functionality

    def value_changed(self):
        """Allow overriding so that another class can add functionality"""
        pass

    def __update_options(self):
        """
        Handle the change of the available options.
        """
        self.__combo.remove_all()
        # self.__combo.get_model().clear()
        cur_val = self.__option.get_value()
        active_index = 0
        current_index = 0
        for value, description in self.__option.get_items():
            self.__combo.append_text(description)
            if value == cur_val:
                active_index = current_index
            current_index += 1
        self.__combo.set_active(active_index)

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        self.__combo.handler_block(self.changekey)
        self.__update_options()
        self.__combo.handler_unblock(self.changekey)

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option.disconnect(self.conkey1)
        self.__option.disconnect(self.conkey2)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiPersonOption class
#
# -------------------------------------------------------------------------
class GuiPersonOption(Gtk.Box):
    """
    This class displays an option that allows a person from the
    database to be selected.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.PersonOption
        @return: nothing
        """
        Gtk.Box.__init__(self)
        self.__option = option
        self.__dbstate = dbstate
        self.__db = dbstate.get_database()
        self.__uistate = uistate
        self.__track = track
        self.__person_label = Gtk.Label()
        self.__person_label.set_halign(Gtk.Align.START)

        pevt = Gtk.EventBox()
        pevt.add(self.__person_label)
        person_button = widgets.SimpleButton("gtk-index", self.__get_person_clicked)
        person_button.set_relief(Gtk.ReliefStyle.NORMAL)

        self.pack_start(pevt, False, True, 0)
        self.pack_end(person_button, False, True, 0)

        gid = self.__option.get_value()

        # Pick up the active person
        try:
            person_handle = self.__uistate.get_active("Person")
            person = self.__dbstate.db.get_person_from_handle(person_handle)

            if override or not person:
                # Pick up the stored option value if there is one
                person = self.__db.get_person_from_gramps_id(gid)

            if not person:
                # If all else fails, get the default person to avoid bad values
                person = self.__db.get_default_person()

            if not person:
                person = self.__db.find_initial_person()
        except HandleError:
            person = None

        self.__update_person(person)

        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.conkey = self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        pevt.set_tooltip_text(self.__option.get_help())
        person_button.set_tooltip_text(_("Select a different person"))

    def __get_person_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the button to choose a different person.
        """
        # Create a filter for the person selector.
        rfilter = GenericFilter()
        rfilter.set_logical_op("or")
        rfilter.add_rule(rules.person.IsBookmarked([]))
        rfilter.add_rule(rules.person.HasIdOf([self.__option.get_value()]))

        # Add the database home person if one exists.
        default_person = self.__db.get_default_person()
        if default_person:
            gid = default_person.get_gramps_id()
            rfilter.add_rule(rules.person.HasIdOf([gid]))

        # Add the selected person if one exists.
        person_handle = self.__uistate.get_active("Person")
        active_person = self.__dbstate.db.get_person_from_handle(person_handle)
        if active_person:
            gid = active_person.get_gramps_id()
            rfilter.add_rule(rules.person.HasIdOf([gid]))

        select_class = SelectorFactory("Person")
        sel = select_class(
            self.__dbstate,
            self.__uistate,
            self.__track,
            title=_("Select a person for the report"),
            filter=rfilter,
        )
        person = sel.run()
        self.__update_person(person)

    def __update_person(self, person):
        """
        Update the currently selected person.
        """
        if person:
            name = _nd.display(person)
            gid = person.get_gramps_id()
            self.__person_label.set_text("%s (%s)" % (name, gid))
            self.__option.set_value(gid)

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        gid = self.__option.get_value()
        name = _nd.display(self.__db.get_person_from_gramps_id(gid))

        self.__person_label.set_text("%s (%s)" % (name, gid))

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option.disconnect(self.conkey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiFamilyOption class
#
# -------------------------------------------------------------------------
class GuiFamilyOption(Gtk.Box):
    """
    This class displays an option that allows a family from the
    database to be selected.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.FamilyOption
        @return: nothing
        """
        Gtk.Box.__init__(self)
        self.__option = option
        self.__dbstate = dbstate
        self.__db = dbstate.get_database()
        self.__uistate = uistate
        self.__track = track
        self.__family_label = Gtk.Label()
        self.__family_label.set_halign(Gtk.Align.START)

        pevt = Gtk.EventBox()
        pevt.add(self.__family_label)
        family_button = widgets.SimpleButton("gtk-index", self.__get_family_clicked)
        family_button.set_relief(Gtk.ReliefStyle.NORMAL)

        self.pack_start(pevt, False, True, 0)
        self.pack_end(family_button, False, True, 0)

        self.__initialize_family(override)

        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.conkey = self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        pevt.set_tooltip_text(self.__option.get_help())
        family_button.set_tooltip_text(_("Select a different family"))

    def __initialize_family(self, override):
        """
        Find a family to initialize the option with. If there is no specified
        family, try to find a family that the user is likely interested in.
        """
        family_list = []

        fid = self.__option.get_value()
        fid_family = self.__db.get_family_from_gramps_id(fid)
        active_family = self.__uistate.get_active("Family")

        if override and fid_family:
            # Use the stored option value if there is one
            family_list = [fid_family.get_handle()]

        if active_family and not family_list:
            # Use the active family if one is selected
            family_list = [active_family]

        if not family_list:
            # Next try the family of the active person
            person_handle = self.__uistate.get_active("Person")
            person = self.__dbstate.db.get_person_from_handle(person_handle)
            if person:
                family_list = person.get_family_handle_list()

        if fid_family and not family_list:
            # Next try the stored option value if there is one
            family_list = [fid_family.get_handle()]

        if not family_list:
            # Next try the family of the default person in the database.
            person = self.__db.get_default_person()
            if person:
                family_list = person.get_family_handle_list()

        if not family_list:
            # Finally, take any family you can find.
            for family in self.__db.iter_family_handles():
                self.__update_family(family)
                break
        else:
            self.__update_family(family_list[0])

    def __get_family_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the button to choose a different family.
        """
        # Create a filter for the person selector.
        rfilter = GenericFilterFactory("Family")()
        rfilter.set_logical_op("or")

        # Add the current family
        rfilter.add_rule(rules.family.HasIdOf([self.__option.get_value()]))

        # Add all bookmarked families
        rfilter.add_rule(rules.family.IsBookmarked([]))

        # Add the families of the database home person if one exists.
        default_person = self.__db.get_default_person()
        if default_person:
            family_list = default_person.get_family_handle_list()
            for family_handle in family_list:
                family = self.__db.get_family_from_handle(family_handle)
                gid = family.get_gramps_id()
                rfilter.add_rule(rules.family.HasIdOf([gid]))

        # Add the families of the selected person if one exists.
        # Same code as above one ! See bug #5032 feature request #5038
        ### active_person = self.__uistate.get_active('Person') ###
        # active_person = self.__db.get_default_person()
        # if active_person:
        # family_list = active_person.get_family_handle_list()
        # for family_handle in family_list:
        # family = self.__db.get_family_from_handle(family_handle)
        # gid = family.get_gramps_id()
        # rfilter.add_rule(rules.family.HasIdOf([gid]))

        select_class = SelectorFactory("Family")
        sel = select_class(self.__dbstate, self.__uistate, self.__track, filter=rfilter)
        family = sel.run()
        if family:
            self.__update_family(family.get_handle())

    def __update_family(self, handle):
        """
        Update the currently selected family.
        """
        if handle:
            family = self.__dbstate.db.get_family_from_handle(handle)
            family_id = family.get_gramps_id()
            fhandle = family.get_father_handle()
            mhandle = family.get_mother_handle()

            if fhandle:
                father = self.__db.get_person_from_handle(fhandle)
                father_name = _nd.display(father)
            else:
                father_name = _("unknown father")

            if mhandle:
                mother = self.__db.get_person_from_handle(mhandle)
                mother_name = _nd.display(mother)
            else:
                mother_name = _("unknown mother")

            name = _("%(father_name)s and %(mother_name)s (%(family_id)s)") % {
                "father_name": father_name,
                "mother_name": mother_name,
                "family_id": family_id,
            }

            self.__family_label.set_text(name)
            self.__option.set_value(family_id)

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        fid = self.__option.get_value()
        handle = self.__db.get_family_from_gramps_id(fid).get_handle()

        # Need to disable signals as __update_family() calls set_value()
        # which would launch the 'value-changed' signal which is what
        # we are reacting to here in the first place (don't need the
        # signal repeated)
        self.__option.disable_signals()
        self.__update_family(handle)
        self.__option.enable_signals()

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option.disconnect(self.conkey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiNoteOption class
#
# -------------------------------------------------------------------------
class GuiNoteOption(Gtk.Box):
    """
    This class displays an option that allows a note from the
    database to be selected.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.NoteOption
        @return: nothing
        """
        Gtk.Box.__init__(self)
        self.__option = option
        self.__dbstate = dbstate
        self.__db = dbstate.get_database()
        self.__uistate = uistate
        self.__track = track
        self.__note_label = Gtk.Label()
        self.__note_label.set_halign(Gtk.Align.START)

        pevt = Gtk.EventBox()
        pevt.add(self.__note_label)
        note_button = widgets.SimpleButton("gtk-index", self.__get_note_clicked)
        note_button.set_relief(Gtk.ReliefStyle.NORMAL)

        self.pack_start(pevt, False, True, 0)
        self.pack_end(note_button, False, True, 0)

        # Initialize to the current value
        nid = self.__option.get_value()
        note = self.__db.get_note_from_gramps_id(nid)
        self.__update_note(note)

        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        pevt.set_tooltip_text(self.__option.get_help())
        note_button.set_tooltip_text(_("Select an existing note"))

    def __get_note_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the button to choose a different note.
        """
        select_class = SelectorFactory("Note")
        sel = select_class(self.__dbstate, self.__uistate, self.__track)
        note = sel.run()
        self.__update_note(note)

    def __update_note(self, note):
        """
        Update the currently selected note.
        """
        if note:
            note_id = note.get_gramps_id()
            txt = " ".join(note.get().split())
            if len(txt) > 35:
                txt = txt[:35] + "..."
            txt = "%s [%s]" % (txt, note_id)

            self.__note_label.set_text(txt)
            self.__option.set_value(note_id)
        else:
            txt = "<i>%s</i>" % _("No note given, click button to select one")
            self.__note_label.set_text(txt)
            self.__note_label.set_use_markup(True)
            self.__option.set_value("")

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        nid = self.__option.get_value()
        note = self.__db.get_note_from_gramps_id(nid)

        # Need to disable signals as __update_note() calls set_value()
        # which would launch the 'value-changed' signal which is what
        # we are reacting to here in the first place (don't need the
        # signal repeated)
        self.__option.disable_signals()
        self.__update_note(note)
        self.__option.enable_signals()

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiMediaOption class
#
# -------------------------------------------------------------------------
class GuiMediaOption(Gtk.Box):
    """
    This class displays an option that allows a media object from the
    database to be selected.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.MediaOption
        @return: nothing
        """
        Gtk.Box.__init__(self)
        self.__option = option
        self.__dbstate = dbstate
        self.__db = dbstate.get_database()
        self.__uistate = uistate
        self.__track = track
        self.__media_label = Gtk.Label()
        self.__media_label.set_halign(Gtk.Align.START)

        pevt = Gtk.EventBox()
        pevt.add(self.__media_label)
        media_button = widgets.SimpleButton("gtk-index", self.__get_media_clicked)
        media_button.set_relief(Gtk.ReliefStyle.NORMAL)

        self.pack_start(pevt, False, True, 0)
        self.pack_end(media_button, False, True, 0)

        # Initialize to the current value
        mid = self.__option.get_value()
        media = self.__db.get_media_from_gramps_id(mid)
        self.__update_media(media)

        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        pevt.set_tooltip_text(self.__option.get_help())
        media_button.set_tooltip_text(_("Select an existing media object"))

    def __get_media_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the button to choose a different note.
        """
        select_class = SelectorFactory("Media")
        sel = select_class(self.__dbstate, self.__uistate, self.__track)
        media = sel.run()
        self.__update_media(media)

    def __update_media(self, media):
        """
        Update the currently selected media.
        """
        if media:
            media_id = media.get_gramps_id()
            txt = "%s [%s]" % (media.get_description(), media_id)

            self.__media_label.set_text(txt)
            self.__option.set_value(media_id)
        else:
            txt = "<i>%s</i>" % _("No image given, click button to select one")
            self.__media_label.set_text(txt)
            self.__media_label.set_use_markup(True)
            self.__option.set_value("")

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        mid = self.__option.get_value()
        media = self.__db.get_media_from_gramps_id(mid)

        # Need to disable signals as __update_media() calls set_value()
        # which would launch the 'value-changed' signal which is what
        # we are reacting to here in the first place (don't need the
        # signal repeated)
        self.__option.disable_signals()
        self.__update_media(media)
        self.__option.enable_signals()

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiPersonListOption class
#
# -------------------------------------------------------------------------
class GuiPersonListOption(Gtk.Box):
    """
    This class displays a widget that allows multiple people from the
    database to be selected.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.PersonListOption
        @return: nothing
        """
        Gtk.Box.__init__(self)
        self.__option = option
        self.__dbstate = dbstate
        self.__db = dbstate.get_database()
        self.__uistate = uistate
        self.__track = track
        self.set_size_request(150, 100)

        self.__model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        self.__tree_view = Gtk.TreeView(model=self.__model)
        col1 = Gtk.TreeViewColumn(_("Name"), Gtk.CellRendererText(), text=0)
        col2 = Gtk.TreeViewColumn(_("ID"), Gtk.CellRendererText(), text=1)
        col1.set_resizable(True)
        col2.set_resizable(True)
        col1.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col2.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col1.set_sort_column_id(0)
        col2.set_sort_column_id(1)
        self.__tree_view.append_column(col1)
        self.__tree_view.append_column(col2)
        self.__scrolled_window = Gtk.ScrolledWindow()
        self.__scrolled_window.add(self.__tree_view)
        self.__scrolled_window.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC
        )
        self.__scrolled_window.set_shadow_type(Gtk.ShadowType.OUT)

        self.pack_start(self.__scrolled_window, True, True, 0)

        self.__value_changed()

        # now setup the '+' and '-' pushbutton for adding/removing people from
        # the container
        self.__add_person = widgets.SimpleButton("list-add", self.__add_person_clicked)
        self.__del_person = widgets.SimpleButton(
            "list-remove", self.__del_person_clicked
        )
        self.__vbbox = Gtk.ButtonBox(orientation=Gtk.Orientation.VERTICAL)
        self.__vbbox.add(self.__add_person)
        self.__vbbox.add(self.__del_person)
        self.__vbbox.set_layout(Gtk.ButtonBoxStyle.SPREAD)
        self.pack_end(self.__vbbox, False, False, 0)

        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.__tree_view.set_tooltip_text(self.__option.get_help())

    def __add_person_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the add person button.
        """
        # people we already have must be excluded
        # so we don't list them multiple times
        skip_list = set()
        tree_iter = self.__model.get_iter_first()
        while tree_iter:
            gid = self.__model.get_value(tree_iter, 1)  # get the GID in col. #1
            person = self.__db.get_person_from_gramps_id(gid)
            skip_list.add(person.get_handle())
            tree_iter = self.__model.iter_next(tree_iter)

        select_class = SelectorFactory("Person")
        sel = select_class(self.__dbstate, self.__uistate, self.__track, skip=skip_list)
        person = sel.run()
        if person:
            name = _nd.display(person)
            gid = person.get_gramps_id()
            self.__model.append([name, gid])

            # if this person has a spouse, ask if we should include the spouse
            # in the list of "people of interest"
            #
            # NOTE: we may want to make this an optional thing, determined
            # by the use of a parameter at the time this class is instatiated
            family_list = person.get_family_handle_list()
            for family_handle in family_list:
                family = self.__db.get_family_from_handle(family_handle)

                if person.get_handle() == family.get_father_handle():
                    spouse_handle = family.get_mother_handle()
                else:
                    spouse_handle = family.get_father_handle()

                if spouse_handle and (spouse_handle not in skip_list):
                    spouse = self.__db.get_person_from_handle(spouse_handle)
                    spouse_name = _nd.display(spouse)
                    text = _("Also include %s?") % spouse_name

                    prompt = OptionDialog(
                        _("Select Person"),
                        text,
                        _("No"),
                        None,
                        _("Yes"),
                        None,
                        parent=self.__uistate.window,
                    )
                    if prompt.get_response() == Gtk.ResponseType.YES:
                        gid = spouse.get_gramps_id()
                        self.__model.append([spouse_name, gid])

            self.__update_value()

    def __del_person_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the delete person button.
        """
        (path, column) = self.__tree_view.get_cursor()
        if path:
            tree_iter = self.__model.get_iter(path)
            self.__model.remove(tree_iter)
            self.__update_value()

    def __update_value(self):
        """
        Parse the object and return.
        """
        gidlist = ""
        tree_iter = self.__model.get_iter_first()
        while tree_iter:
            gid = self.__model.get_value(tree_iter, 1)
            gidlist = gidlist + gid + " "
            tree_iter = self.__model.iter_next(tree_iter)

        # Supress signals so that the set_value() handler
        # (__value_changed()) doesn't get called
        self.__option.disable_signals()
        self.__option.set_value(gidlist)
        self.__option.enable_signals()

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        value = self.__option.get_value()

        if not isinstance(value, str):
            # Convert array into a string
            # (convienence so that programmers can
            # set value using a list)
            value = " ".join(value)

            # Need to change __option value to be the string

            self.__option.disable_signals()
            self.__option.set_value(value)
            self.__option.enable_signals()

        # Remove all entries (the new values will REPLACE
        # rather than APPEND)
        self.__model.clear()

        for gid in value.split():
            person = self.__db.get_person_from_gramps_id(gid)
            if person:
                name = _nd.display(person)
                self.__model.append([name, gid])

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiPlaceListOption class
#
# -------------------------------------------------------------------------
class GuiPlaceListOption(Gtk.Box):
    """
    This class displays a widget that allows multiple places from the
    database to be selected.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.PlaceListOption
        @return: nothing
        """
        Gtk.Box.__init__(self)
        self.__option = option
        self.__dbstate = dbstate
        self.__db = dbstate.get_database()
        self.__uistate = uistate
        self.__track = track
        self.set_size_request(150, 150)

        self.__model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        self.__tree_view = Gtk.TreeView(self.__model)
        col1 = Gtk.TreeViewColumn(_("Place"), Gtk.CellRendererText(), text=0)
        col2 = Gtk.TreeViewColumn(_("ID"), Gtk.CellRendererText(), text=1)
        col1.set_resizable(True)
        col2.set_resizable(True)
        col1.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col2.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col1.set_sort_column_id(0)
        col2.set_sort_column_id(1)
        self.__tree_view.append_column(col1)
        self.__tree_view.append_column(col2)
        self.__scrolled_window = Gtk.ScrolledWindow()
        self.__scrolled_window.add(self.__tree_view)
        self.__scrolled_window.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC
        )
        self.__scrolled_window.set_shadow_type(Gtk.ShadowType.OUT)

        self.pack_start(self.__scrolled_window, True, True, 0)

        self.__value_changed()

        # now setup the '+' and '-' pushbutton for adding/removing places from
        # the container
        self.__add_place = widgets.SimpleButton("list-add", self.__add_place_clicked)
        self.__del_place = widgets.SimpleButton("list-remove", self.__del_place_clicked)
        self.__vbbox = Gtk.ButtonBox(orientation=Gtk.Orientation.VERTICAL)
        self.__vbbox.add(self.__add_place)
        self.__vbbox.add(self.__del_place)
        self.__vbbox.set_layout(Gtk.ButtonBoxStyle.SPREAD)
        self.pack_end(self.__vbbox, False, False, 0)

        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.__tree_view.set_tooltip_text(self.__option.get_help())

    def __add_place_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the add place button.
        """
        # places we already have must be excluded
        # so we don't list them multiple times
        skip_list = set()
        tree_iter = self.__model.get_iter_first()
        while tree_iter:
            gid = self.__model.get_value(tree_iter, 1)  # get the GID in col. #1
            place = self.__db.get_place_from_gramps_id(gid)
            skip_list.add(place.get_handle())
            tree_iter = self.__model.iter_next(tree_iter)

        select_class = SelectorFactory("Place")
        sel = select_class(self.__dbstate, self.__uistate, self.__track, skip=skip_list)
        place = sel.run()
        if place:
            place_name = _pd.display(self.__db, place)
            gid = place.get_gramps_id()
            self.__model.append([place_name, gid])
            self.__update_value()

    def __del_place_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the delete place button.
        """
        (path, column) = self.__tree_view.get_cursor()
        if path:
            tree_iter = self.__model.get_iter(path)
            self.__model.remove(tree_iter)
            self.__update_value()

    def __update_value(self):
        """
        Parse the object and return.
        """
        gidlist = ""
        tree_iter = self.__model.get_iter_first()
        while tree_iter:
            gid = self.__model.get_value(tree_iter, 1)
            gidlist = gidlist + gid + " "
            tree_iter = self.__model.iter_next(tree_iter)
        self.__option.set_value(gidlist)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        value = self.__option.get_value()

        if not isinstance(value, str):
            # Convert array into a string
            # (convienence so that programmers can
            # set value using a list)
            value = " ".join(value)

            # Need to change __option value to be the string

            self.__option.disable_signals()
            self.__option.set_value(value)
            self.__option.enable_signals()

        # Remove all entries (the new values will REPLACE
        # rather than APPEND)
        self.__model.clear()

        for gid in value.split():
            place = self.__db.get_place_from_gramps_id(gid)
            if place:
                place_name = _pd.display(self.__db, place)
                self.__model.append([place_name, gid])

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiSurnameColorOption class
#
# -------------------------------------------------------------------------
class GuiSurnameColorOption(Gtk.Box):
    """
    This class displays a widget that allows multiple surnames to be
    selected from the database, and to assign a colour (not necessarily
    unique) to each one.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.SurnameColorOption
        @return: nothing
        """
        Gtk.Box.__init__(self)
        self.__option = option
        self.__dbstate = dbstate
        self.__db = dbstate.get_database()
        self.__uistate = uistate
        self.__track = track
        item = uistate.gwm.get_item_from_track(track)
        self.__parent = item[0].window if isinstance(item, list) else item.window

        self.set_size_request(150, 150)

        # This will get populated the first time the dialog is run,
        # and used each time after.
        self.__surnames = {}  # list of surnames and count

        self.__model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        self.__tree_view = Gtk.TreeView(model=self.__model)
        self.__tree_view.connect("row-activated", self.__row_clicked)
        col1 = Gtk.TreeViewColumn(_("Surname"), Gtk.CellRendererText(), text=0)
        col2 = Gtk.TreeViewColumn(_("Color"), Gtk.CellRendererText(), text=1)
        col1.set_resizable(True)
        col2.set_resizable(True)
        col1.set_sort_column_id(0)
        col1.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col2.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self.__tree_view.append_column(col1)
        self.__tree_view.append_column(col2)
        self.scrolled_window = Gtk.ScrolledWindow()
        self.scrolled_window.add(self.__tree_view)
        self.scrolled_window.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC
        )
        self.scrolled_window.set_shadow_type(Gtk.ShadowType.OUT)
        self.pack_start(self.scrolled_window, True, True, 0)

        self.add_surname = widgets.SimpleButton("list-add", self.__add_clicked)
        self.del_surname = widgets.SimpleButton("list-remove", self.__del_clicked)
        self.vbbox = Gtk.ButtonBox(orientation=Gtk.Orientation.VERTICAL)
        self.vbbox.add(self.add_surname)
        self.vbbox.add(self.del_surname)
        self.vbbox.set_layout(Gtk.ButtonBoxStyle.SPREAD)
        self.pack_end(self.vbbox, False, False, 0)

        self.__value_changed()

        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.__tree_view.set_tooltip_text(self.__option.get_help())

    def __add_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the add surname button.
        """
        skip_list = set()
        tree_iter = self.__model.get_iter_first()
        while tree_iter:
            surname = self.__model.get_value(tree_iter, 0)
            skip_list.add(surname.encode("iso-8859-1", "xmlcharrefreplace"))
            tree_iter = self.__model.iter_next(tree_iter)

        ln_dialog = LastNameDialog(
            self.__db, self.__uistate, self.__track, self.__surnames, skip_list
        )
        surname_set = ln_dialog.run()
        for surname in surname_set:
            self.__model.append([surname, "#ffffff"])
        self.__update_value()

    def __del_clicked(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the delete surname button.
        """
        (path, column) = self.__tree_view.get_cursor()
        if path:
            tree_iter = self.__model.get_iter(path)
            self.__model.remove(tree_iter)
            self.__update_value()

    def __row_clicked(self, treeview, path, column):
        """
        Handle the case of a row being clicked on.
        """
        # get the surname and colour value for this family
        tree_iter = self.__model.get_iter(path)
        surname = self.__model.get_value(tree_iter, 0)
        rgba = Gdk.RGBA()
        rgba.parse(self.__model.get_value(tree_iter, 1))

        title = _("Select color for %s") % surname
        colour_dialog = Gtk.ColorChooserDialog(title=title, transient_for=self.__parent)
        colour_dialog.set_rgba(rgba)
        response = colour_dialog.run()

        if response == Gtk.ResponseType.OK:
            rgba = colour_dialog.get_rgba()
            colour_name = "#%02x%02x%02x" % (
                int(rgba.red * 255),
                int(rgba.green * 255),
                int(rgba.blue * 255),
            )
            self.__model.set_value(tree_iter, 1, colour_name)

        colour_dialog.destroy()
        self.__update_value()

    def __update_value(self):
        """
        Parse the object and return.
        """
        surname_colours = ""
        tree_iter = self.__model.get_iter_first()
        while tree_iter:
            surname = self.__model.get_value(tree_iter, 0)
            # surname = surname.encode('iso-8859-1', 'xmlcharrefreplace')
            colour = self.__model.get_value(tree_iter, 1)
            # Tried to use a dictionary, and tried to save it as a tuple,
            # but couldn't get this to work right -- this is lame, but now
            # the surnames and colours are saved as a plain text string
            #
            # Hmmm...putting whitespace between the fields causes
            # problems when the surname has whitespace -- for example,
            # with surnames like "Del Monte".  So now we insert a non-
            # whitespace character which is unlikely to appear in
            # a surname.  (See bug report #2162.)
            surname_colours += surname + "\xb0" + colour + "\xb0"
            tree_iter = self.__model.iter_next(tree_iter)
        self.__option.set_value(surname_colours)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        value = self.__option.get_value()

        if not isinstance(value, str):
            # Convert dictionary into a string
            # (convienence so that programmers can
            # set value using a dictionary)
            value_str = ""

            for name in value:
                value_str += "%s\xb0%s\xb0" % (name, value[name])

            value = value_str

            # Need to change __option value to be the string

            self.__option.disable_signals()
            self.__option.set_value(value)
            self.__option.enable_signals()

        # Remove all entries (the new values will REPLACE
        # rather than APPEND)
        self.__model.clear()

        # populate the surname/colour treeview
        #
        # For versions prior to 3.0.2, the fields were delimited with
        # whitespace.  However, this causes problems when the surname
        # also has a space within it.  When populating the control,
        # support both the new and old format -- look for the \xb0
        # delimiter, and if it isn't there, assume this is the old-
        # style space-delimited format.  (Bug #2162.)
        if value.find("\xb0") >= 0:
            tmp = value.split("\xb0")
        else:
            tmp = value.split(" ")
        while len(tmp) > 1:
            surname = tmp.pop(0)
            colour = tmp.pop(0)
            self.__model.append([surname, colour])

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiDestinationOption class
#
# -------------------------------------------------------------------------
class GuiDestinationOption(Gtk.Box):
    """
    This class displays an option that allows the user to select a
    DestinationOption.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.DestinationOption
        @return: nothing
        """
        Gtk.Box.__init__(self)
        self.__option = option
        self.__uistate = uistate
        self.__entry = Gtk.Entry()
        self.__entry.set_text(self.__option.get_value())

        self.__button = Gtk.Button()
        img = Gtk.Image()
        img.set_from_icon_name("document-open", Gtk.IconSize.BUTTON)
        self.__button.add(img)
        self.__button.connect("clicked", self.__select_file)

        self.pack_start(self.__entry, True, True, 0)
        self.pack_end(self.__button, False, False, 0)

        # Set up signal handlers when the widget value is changed
        # from user interaction or programmatically.  When handling
        # a specific signal, we need to temporarily block the signal
        # that would call the other signal handler.
        self.changekey = self.__entry.connect("changed", self.__text_changed)
        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.conkey1 = self.__option.connect("options-changed", self.__option_changed)
        self.conkey2 = self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        self.set_tooltip_text(self.__option.get_help())

    def __option_changed(self):
        """
        Handle a change of the option.
        """
        extension = self.__option.get_extension()
        directory = self.__option.get_directory_entry()
        value = self.__option.get_value()

        if not directory and not value.endswith(extension):
            value = value + extension
            self.__option.set_value(value)
        elif directory and value.endswith(extension):
            value = value[: -len(extension)]
            self.__option.set_value(value)

        self.__entry.set_text(self.__option.get_value())

    def __select_file(self, obj):
        """
        Handle the user's request to select a file (or directory).
        """
        if self.__option.get_directory_entry():
            my_action = Gtk.FileChooserAction.SELECT_FOLDER
        else:
            my_action = Gtk.FileChooserAction.SAVE

        fcd = Gtk.FileChooserDialog(
            title=_("Save As"), action=my_action, transient_for=self.__uistate.window
        )
        fcd.add_buttons(
            _("_Cancel"), Gtk.ResponseType.CANCEL, _("_Open"), Gtk.ResponseType.OK
        )

        name = os.path.abspath(self.__option.get_value())
        if self.__option.get_directory_entry():
            while not os.path.isdir(name):
                # Keep looking up levels to find a valid drive.
                name, tail = os.path.split(name)
                if not name:
                    # Avoid infinite loops
                    name = get_curr_dir
            fcd.set_current_folder(name)
        else:
            fcd.set_current_name(os.path.basename(name))
            fcd.set_current_folder(os.path.dirname(name))

        status = fcd.run()
        if status == Gtk.ResponseType.OK:
            path = fcd.get_filename()
            if path:
                if not self.__option.get_directory_entry() and not path.endswith(
                    self.__option.get_extension()
                ):
                    path = path + self.__option.get_extension()
                self.__entry.set_text(path)
                self.__option.set_value(path)
        fcd.destroy()

    def __text_changed(self, obj):  # IGNORE:W0613 - obj is unused
        """
        Handle the change of the value made by the user.
        """
        self.__option.disable_signals()
        self.__option.set_value(self.__entry.get_text())
        self.__option.enable_signals()

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        self.__entry.handler_block(self.changekey)
        self.__entry.set_text(self.__option.get_value())
        self.__entry.handler_unblock(self.changekey)

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option.disconnect(self.conkey1)
        self.__option.disconnect(self.conkey2)
        self.__option = None


# -------------------------------------------------------------------------
#
# GuiStyleOption class
#
# -------------------------------------------------------------------------
class GuiStyleOption(GuiEnumeratedListOption):  # TODO this is likely dead code
    """
    This class displays a StyleOption.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        """
        @param option: The option to display.
        @type option: gen.plug.menu.StyleOption
        @return: nothing
        """
        GuiEnumeratedListOption.__init__(self, option, dbstate, uistate, track)
        self.__option = option

        self.__button = Gtk.Button("%s..." % _("Style Editor"))
        self.__button.connect("clicked", self.__on_style_edit_clicked)

        self.pack_end(self.__button, False, False)
        self.uistate = uistate
        self.track = track

    def __on_style_edit_clicked(self, *obj):
        """The user has clicked on the 'Edit Styles' button.  Create a
        style sheet editor object and let them play.  When they are
        done, update the displayed styles."""
        from gramps.gen.plug.docgen import StyleSheetList
        from .report._styleeditor import StyleListDisplay

        style_list = StyleSheetList(
            self.__option.get_style_file(), self.__option.get_default_style()
        )
        StyleListDisplay(style_list, self.uistate, self.track)

        new_items = []
        for style_name in style_list.get_style_names():
            new_items.append((style_name, style_name))
        self.__option.set_items(new_items)


# -------------------------------------------------------------------------
#
# GuiBooleanListOption class
#
# -------------------------------------------------------------------------
class GuiBooleanListOption(Gtk.Box):
    """
    This class displays an option that provides a list of check boxes.
    Each possible value is assigned a value and a description.
    """

    def __init__(self, option, dbstate, uistate, track, override):
        Gtk.Box.__init__(self)
        self.__option = option
        self.__cbutton = []

        default = option.get_value().split(",")
        if len(default) < 15:
            columns = 2  # number of checkbox columns
        else:
            columns = 3
        column = []
        for dummy in range(columns):
            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            self.pack_start(vbox, True, True, 0)
            column.append(vbox)
            vbox.show()

        counter = 0
        this_column_counter = 0
        ncolumn = 0
        for description in option.get_descriptions():
            button = Gtk.CheckButton(label=description)
            self.__cbutton.append(button)
            if counter < len(default):
                if default[counter] == "True":
                    button.set_active(True)
            button.connect("toggled", self.__list_changed)
            # show the items vertically, not alternating left and right
            # (if the number is uneven, the left column(s) will have one more)
            column[ncolumn].pack_start(button, True, True, 0)
            button.show()
            counter += 1
            this_column_counter += 1
            this_column_gets = (len(default) + (columns - (ncolumn + 1))) // columns
            if this_column_counter + 1 > this_column_gets:
                ncolumn += 1
                this_column_counter = 0

        self.valuekey = self.__option.connect("value-changed", self.__value_changed)

        self.__option.connect("avail-changed", self.__update_avail)
        self.__update_avail()

        self.set_tooltip_text(self.__option.get_help())

    def __list_changed(self, button):
        """
        Handle the change of the value made by the user.
        """
        value = ""
        for button in self.__cbutton:
            value = value + str(button.get_active()) + ","
        value = value[: len(value) - 1]

        self.__option.disable_signals()
        self.__option.set_value(value)
        self.__option.enable_signals()

    def __update_avail(self):
        """
        Update the availability (sensitivity) of this widget.
        """
        avail = self.__option.get_available()
        self.set_sensitive(avail)

    def __value_changed(self):
        """
        Handle the change made programmatically
        """
        value = self.__option.get_value()

        self.__option.disable_signals()

        for button in self.__cbutton:
            for key in value:
                if key == button.get_label():
                    bool_value = value[key] == "True" or value[key] == True
                    button.set_active(bool_value)

        # Update __option value so that it's correct
        self.__list_changed(None)

        self.__option.enable_signals()

    def clean_up(self):
        """
        remove stuff that blocks garbage collection
        """
        self.__option.disconnect(self.valuekey)
        self.__option = None


# -----------------------------------------------------------------------------#
#                                                                             #
#   Table mapping menu types to gui widgets used in make_gui_option function  #
#                                                                             #
# -----------------------------------------------------------------------------#

from gramps.gen.plug import menu as menu

_OPTIONS = (
    (menu.BooleanListOption, True, GuiBooleanListOption),
    (menu.BooleanOption, False, GuiBooleanOption),
    (menu.ColorOption, True, GuiColorOption),
    (menu.DestinationOption, True, GuiDestinationOption),
    (menu.EnumeratedListOption, True, GuiEnumeratedListOption),
    (menu.FamilyOption, True, GuiFamilyOption),
    (menu.MediaOption, True, GuiMediaOption),
    (menu.NoteOption, True, GuiNoteOption),
    (menu.NumberOption, True, GuiNumberOption),
    (menu.PersonListOption, True, GuiPersonListOption),
    (menu.PersonOption, True, GuiPersonOption),
    (menu.PlaceListOption, True, GuiPlaceListOption),
    (menu.StringOption, True, GuiStringOption),
    (menu.StyleOption, True, GuiStyleOption),
    (menu.SurnameColorOption, True, GuiSurnameColorOption),
    (menu.TextOption, True, GuiTextOption),
    # This entry must be last!
    (menu.Option, None, None),
)
del menu


def make_gui_option(option, dbstate, uistate, track, override=False):
    """
    Stand-alone function so that Options can be used in other
    ways, too. Takes an Option and returns a GuiOption.

    override: if True will override the GuiOption's normal behavior
        (in a GuiOption-dependant fashion, for instance in a GuiPersonOption
        it will force the use of the options's value to set the GuiOption)
    """

    label, widget = True, None
    pmgr = GuiPluginManager.get_instance()
    external_options = pmgr.get_external_opt_dict()
    if option.__class__ in external_options:
        widget = external_options[option.__class__]
    else:
        for type_, label, widget in _OPTIONS:
            if isinstance(option, type_):
                break
        else:
            raise AttributeError(
                "can't make GuiOption: unknown option type: '%s'" % option
            )

    if widget:
        widget = widget(option, dbstate, uistate, track, override)

    return widget, label


def add_gui_options(dialog):
    """
    Stand-alone function to add user options to the GUI.
    """
    if not hasattr(dialog.options, "menu"):
        return
    o_menu = dialog.options.menu
    options_dict = dialog.options.options_dict
    for category in o_menu.get_categories():
        for name in o_menu.get_option_names(category):
            option = o_menu.get_option(category, name)

            # override option default with xml-saved value:
            if name in options_dict:
                option.set_value(options_dict[name])

            widget, label = make_gui_option(
                option, dialog.dbstate, dialog.uistate, dialog.track
            )
            if widget is not None:
                if label:
                    dialog.add_frame_option(category, option.get_label(), widget)
                else:
                    dialog.add_frame_option(category, "", widget)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009-2011  Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ...widgets import SimpleButton
from .grampstab import GrampsTab
from gramps.gen.errors import WindowActiveError
from ...utils import match_primary_mask

_KP_ENTER = Gdk.keyval_from_name("KP_Enter")
_RETURN = Gdk.keyval_from_name("Return")
_DEL = Gdk.keyval_from_name("Delete")
_ADD = Gdk.keyval_from_name("Insert")
_OPEN = Gdk.keyval_from_name("o")
_LEFT = Gdk.keyval_from_name("Left")
_RIGHT = Gdk.keyval_from_name("Right")


# -------------------------------------------------------------------------
#
# Classes
#
# -------------------------------------------------------------------------
class ButtonTab(GrampsTab):
    """
    This class derives from the base GrampsTab, yet is not a usable Tab. It
    serves as another base tab for classes which need an Add/Edit/Remove button
    combination.
    """

    _MSG = {
        "add": _("Add"),
        "del": _("Remove"),
        "edit": _("Edit"),
        "share": _("Share"),
        "jump": _("Jump To"),
        "up": _("Move Up"),
        "down": _("Move Down"),
        "left": _("Move Left"),
        "right": _("Move right"),
    }
    L_R = 2  # indicator for left/right move buttons

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        name,
        share_button=False,
        move_buttons=False,
        jump_button=False,
        top_label=None,
    ):
        """
        Similar to the base class, except after Build.

        @param dbstate: The database state. Contains a reference to
        the database, along with other state information. The GrampsTab
        uses this to access the database and to pass to and created
        child windows (such as edit dialogs).
        @type dbstate: DbState
        @param uistate: The UI state. Used primarily to pass to any created
        subwindows.
        @type uistate: DisplayState
        @param track: The window tracking mechanism used to manage windows.
        This is only used to pass to generted child windows.
        @type track: list
        @param name: Notebook label name
        @type name: str/unicode
        @param share_button: Add a share button to the Notebook tab or not
        @type name: bool
        @param move_buttons: Add up and down button to the Notebook tab or not
        @type name: bool
        @param jump_button: Add a goto button
        @type name: bool
        @param top_label: Add a label in front of the buttons if given
        @type top_label: string or None for no label
        """
        self.dirty_selection = False
        GrampsTab.__init__(self, dbstate, uistate, track, name)
        self._create_buttons(share_button, move_buttons, jump_button, top_label)

    def _create_buttons(self, share_button, move_buttons, jump_button, top_label):
        """
        Create a button box consisting of three buttons, one for Add,
        one for Edit, and one for Delete.

        Add buttons for Share, Move and Jump depending on parameters. This
        button box is then appended hbox (self).
        Prepend a label if top_label given

        Note: some ButtonTab subclasses override this method.
        """
        if top_label:
            self.top_label = Gtk.Label(label=top_label)
            self.top_label.set_use_markup(True)
            self.track_ref_for_deletion("top_label")

        self.add_btn = SimpleButton("list-add", self.add_button_clicked)
        self.edit_btn = SimpleButton("gtk-edit", self.edit_button_clicked)
        self.del_btn = SimpleButton("list-remove", self.del_button_clicked)
        self.track_ref_for_deletion("add_btn")
        self.track_ref_for_deletion("edit_btn")
        self.track_ref_for_deletion("del_btn")

        self.add_btn.set_tooltip_text(self._MSG["add"])
        self.edit_btn.set_tooltip_text(self._MSG["edit"])
        self.del_btn.set_tooltip_text(self._MSG["del"])

        if share_button:
            self.share_btn = SimpleButton("gtk-index", self.share_button_clicked)
            self.share_btn.set_tooltip_text(self._MSG["share"])
            self.track_ref_for_deletion("share_btn")
        else:
            self.share_btn = None

        if move_buttons:
            l_r = move_buttons == self.L_R
            self.up_btn = SimpleButton(
                "go-previous" if l_r else "go-up", self.up_button_clicked
            )
            self.up_btn.set_tooltip_text(self._MSG["left" if l_r else "up"])
            self.down_btn = SimpleButton(
                "go-next" if l_r else "go-down", self.down_button_clicked
            )
            self.down_btn.set_tooltip_text(self._MSG["right" if l_r else "down"])
            self.track_ref_for_deletion("up_btn")
            self.track_ref_for_deletion("down_btn")
        else:
            self.up_btn = None
            self.down_btn = None

        if jump_button:
            self.jump_btn = SimpleButton("go-jump", self.jump_button_clicked)
            self.track_ref_for_deletion("jump_btn")
            self.jump_btn.set_tooltip_text(self._MSG["jump"])
        else:
            self.jump_btn = None

        hbox = Gtk.Box()
        hbox.set_spacing(6)
        if top_label:
            hbox.pack_start(self.top_label, False, True, 0)
        hbox.pack_start(self.add_btn, False, True, 0)
        if share_button:
            hbox.pack_start(self.share_btn, False, True, 0)
        hbox.pack_start(self.edit_btn, False, True, 0)
        hbox.pack_start(self.del_btn, False, True, 0)
        if move_buttons:
            hbox.pack_start(self.up_btn, False, True, 0)
            hbox.pack_start(self.down_btn, False, True, 0)

        if self.jump_btn:
            hbox.pack_start(self.jump_btn, False, True, 0)
        hbox.show_all()
        self.pack_start(hbox, False, True, 0)

        if self.dbstate.db.readonly:
            self.add_btn.set_sensitive(False)
            self.del_btn.set_sensitive(False)
            if share_button:
                self.share_btn.set_sensitive(False)
            if jump_button and self.jump_btn:
                self.jump_btn.set_sensitive(False)
            if move_buttons:
                self.up_btn.set_sensitive(False)
                self.down_btn.set_sensitive(False)

    def double_click(self, obj, event):
        """
        Handles the double click on list. If the double click occurs,
        the Edit button handler is called
        """
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            try:
                self.edit_button_clicked(obj)
            except WindowActiveError:
                pass

    def key_pressed(self, obj, event):
        """
        Handles the return key being pressed on list. If the key is pressed,
        the Edit button handler is called
        """
        if event.type == Gdk.EventType.KEY_PRESS:
            # print 'key pressed', event.keyval, event.get_state(), _ADD
            if event.keyval in (_RETURN, _KP_ENTER):
                try:
                    self.edit_button_clicked(obj)
                except WindowActiveError:
                    pass
            elif event.keyval in (_DEL,) and self.del_btn:
                if self.dirty_selection or self.dbstate.db.readonly:
                    return
                self.del_button_clicked(obj)
            elif event.keyval in (_ADD,) and self.add_btn:
                if self.dirty_selection or self.dbstate.db.readonly:
                    return
                self.add_button_clicked(obj)
            elif (
                event.keyval in (_OPEN,)
                and self.share_btn
                and match_primary_mask(event.get_state())
            ):
                self.share_button_clicked(obj)
            elif event.keyval in (_LEFT,) and (
                event.get_state() & Gdk.ModifierType.MOD1_MASK
            ):
                self.prev_page()
            elif event.keyval in (_RIGHT,) and (
                event.get_state() & Gdk.ModifierType.MOD1_MASK
            ):
                self.next_page()
            else:
                return
            return True

    def add_button_clicked(self, obj):
        """
        Function called with the Add button is clicked. This function
        should be overridden by the derived class.
        """
        print("Uncaught Add clicked")

    def share_button_clicked(self, obj):
        """
        Function called with the Share button is clicked. This function
        should be overridden by the derived class.
        """
        print("Uncaught Share clicked")

    def jump_button_clicked(self, obj):
        """
        Function called with the Jump button is clicked. This function
        should be overridden by the derived class.
        """
        print("Uncaught Jump clicked")

    def del_button_clicked(self, obj):
        """
        Function called with the Delete button is clicked. This function
        should be overridden by the derived class.
        """
        print("Uncaught Delete clicked")

    def edit_button_clicked(self, obj):
        """
        Function called with the Edit button is clicked or the double
        click is caught. This function should be overridden by the derived
        class.
        """
        print("Uncaught Edit clicked")

    def up_button_clicked(self, obj):
        """
        Function called with the Up button is clicked.
        This function should be overridden by the derived class.
        """
        print("Uncaught Up clicked")

    def down_button_clicked(self, obj):
        """
        Function called with the Down button is clicked.
        This function should be overridden by the derived class.
        """
        print("Uncaught Down clicked")

    def _selection_changed(self, obj=None):
        """
        Attached to the selection's 'changed' signal. Checks
        to see if anything is selected. If it is, the edit and
        delete buttons are enabled, otherwise the are disabled.
        """
        # Comparing to None is important, as empty strings
        # and 0 can be returned
        # This method is called as callback on change, and can be called
        # explicitly, dirty_selection must make sure they do not interact
        if self.dirty_selection:
            return
        if self.get_selected() is not None:
            self.edit_btn.set_sensitive(True)
            if self.jump_btn:
                self.jump_btn.set_sensitive(True)
            if not self.dbstate.db.readonly:
                self.del_btn.set_sensitive(True)
            # note: up and down cannot be set unsensitive after clicked
            #       or they do not respond to a next click
            # if self.up_btn :
            #    self.up_btn.set_sensitive(True)
            #    self.down_btn.set_sensitive(True)
        else:
            self.edit_btn.set_sensitive(False)
            if self.jump_btn:
                self.jump_btn.set_sensitive(False)
            if not self.dbstate.db.readonly:
                self.del_btn.set_sensitive(False)
            # note: up and down cannot be set unsensitive after clicked
            #       or they do not respond to a next click
            # if self.up_btn :
            #    self.up_btn.set_sensitive(False)
            #    self.down_btn.set_sensitive(False)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Database Processing/Fix Capitalization of Family Names"""

# -------------------------------------------------------------------------
#
# gnome/gtk
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.db import find_surname_name, DbTxn
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gui.utils import ProgressMeter
from gramps.gui.display import display_help
from gramps.gui.managedwindow import ManagedWindow

from gramps.gui.dialog import OkDialog
from gramps.gui.plug import tool
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gui.glade import Glade

# -------------------------------------------------------------------------
#
# constants
#
# -------------------------------------------------------------------------

prefix_list = [
    "de",
    "van",
    "von",
    "di",
    "le",
    "du",
    "dela",
    "della",
    "des",
    "vande",
    "ten",
    "da",
    "af",
    "den",
    "das",
    "dello",
    "del",
    "en",
    "ein",
    "el" "et",
    "les",
    "lo",
    "los",
    "un",
    "um",
    "una",
    "uno",
]

WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Fix_Capitalization_of_Family_Names", "manual")


# -------------------------------------------------------------------------
#
# ChangeNames
#
# -------------------------------------------------------------------------
class ChangeNames(tool.BatchTool, ManagedWindow):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        self.label = _("Capitalization changes")
        self.cb = callback

        ManagedWindow.__init__(self, uistate, [], self.__class__)
        self.set_window(Gtk.Window(), Gtk.Label(), "")

        tool.BatchTool.__init__(self, dbstate, user, options_class, name)
        if self.fail:
            return

        self.progress = ProgressMeter(
            _("Checking Family Names"), "", parent=uistate.window
        )
        self.progress.set_pass(
            _("Searching family names"), len(self.db.get_surname_list())
        )
        self.name_list = []

        for name in self.db.get_surname_list():
            name.strip()
            namesplitSP = name.split()
            lSP = len(namesplitSP)
            namesplitHY = name.split("-")
            lHY = len(namesplitHY)
            if lSP == lHY == 1:
                if name != name.capitalize():
                    # Single surname without hyphen(s)
                    self.name_list.append(name)
            # if lSP == 1 and lHY > 1:
            # print "LSP==1", name, name.capitalize()
            # if name != name.capitalize():
            # Single surname with hyphen(s)
            # self.name_list.append(name)
            if lSP > 1 and lHY == 1:
                # more than one string in surname but no hyphen
                # check if first string is in prefix_list, if so test for cap in rest
                s1 = 0
                if namesplitSP[0].lower() in prefix_list:
                    s1 = 1
                for x in range(len(namesplitSP) - s1):
                    # check if any subsurname is not cap
                    notcap = False
                    if namesplitSP[s1 + x] != namesplitSP[s1 + x].capitalize():
                        notcap = True
                        break
                if notcap:
                    # Multiple surnames possibly after prefix
                    self.name_list.append(name)
            if lHY > 1:
                # more than one string in surname but hyphen(s) exists
                # check if first string is in prefix_list, if so test for cap
                if namesplitSP[0].lower() in prefix_list:
                    namesplitHY[0] = namesplitHY[0].replace(namesplitSP[0], "").strip()
                for x in range(len(namesplitHY)):
                    # check if any subsurname is not cap
                    notcap = False
                    if namesplitHY[x] != namesplitHY[x].capitalize():
                        notcap = True
                        break
                if notcap:
                    # Multiple surnames possibly after frefix
                    self.name_list.append(name)

            if uistate:
                self.progress.step()

        if self.name_list:
            self.display()
        else:
            self.progress.close()
            self.close()
            OkDialog(
                _("No modifications made"),
                _("No capitalization changes were detected."),
                parent=uistate.window,
            )

    def name_cap(self, name):
        name.strip()
        namesplitSP = name.split()
        lSP = len(namesplitSP)
        lHY = len(name.split("-"))
        namesep = " "
        if lHY > 1:
            namesep = "-"
            namesplitSP = name.replace(namesep, " ").split()
            lSP = len(namesplitSP)
        if lSP == lHY == 1:
            # if name != name.capitalize():
            # Single surname without space(s) or hyphen(s), normal case
            return name.capitalize()
        else:
            # more than one string in surname but no hyphen
            # check if first string is in prefix_list, if so CAP the rest
            # Names like (von) Kohl(-)Brandt
            result = ""
            s1 = 0
            if namesplitSP[0].lower() in prefix_list:
                s1 = 1
                result = namesplitSP[0].lower() + " "
            for x in range(lSP - s1):
                # CAP all subsurnames
                result = result + namesplitSP[s1 + x].capitalize() + namesep
            return result[:-1]

    def display(self):
        self.top = Glade()
        window = self.top.toplevel
        self.top.connect_signals(
            {
                "destroy_passed_object": self.close,
                "on_ok_clicked": self.on_ok_clicked,
                "on_help_clicked": self.on_help_clicked,
                "on_delete_event": self.close,
            }
        )

        self.list = self.top.get_object("list")
        self.set_window(window, self.top.get_object("title"), self.label)
        self.setup_configs("interface.changenames", 500, 450)

        self.model = Gtk.ListStore(
            GObject.TYPE_BOOLEAN, GObject.TYPE_STRING, GObject.TYPE_STRING
        )

        r = Gtk.CellRendererToggle()
        r.connect("toggled", self.toggled)
        c = Gtk.TreeViewColumn(_("Select"), r, active=0)
        self.list.append_column(c)

        c = Gtk.TreeViewColumn(_("Original Name"), Gtk.CellRendererText(), text=1)
        self.list.append_column(c)

        c = Gtk.TreeViewColumn(
            _("Capitalization Change"), Gtk.CellRendererText(), text=2
        )
        self.list.append_column(c)

        self.list.set_model(self.model)

        self.iter_list = []
        self.progress.set_pass(_("Building display"), len(self.name_list))
        for name in self.name_list:
            handle = self.model.append()
            self.model.set_value(handle, 0, True)
            self.model.set_value(handle, 1, name)
            namecap = self.name_cap(name)
            self.model.set_value(handle, 2, namecap)
            self.iter_list.append(handle)
            self.progress.step()
        self.progress.close()

        self.show()

    def toggled(self, cell, path_string):
        path = tuple(map(int, path_string.split(":")))
        row = self.model[path]
        row[0] = not row[0]

    def build_menu_names(self, obj):
        return (self.label, None)

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(WIKI_HELP_PAGE, WIKI_HELP_SEC)

    def on_ok_clicked(self, obj):
        with DbTxn(_("Capitalization changes"), self.db, batch=True) as self.trans:
            self.db.disable_signals()
            changelist = set(
                self.model.get_value(node, 1)
                for node in self.iter_list
                if self.model.get_value(node, 0)
            )

            # with self.db.get_person_cursor(update=True, commit=True) as cursor:
            #  for handle, data in cursor:
            for handle in self.db.get_person_handles(False):
                person = self.db.get_person_from_handle(handle)
                # person = Person(data)
                change = False
                for name in [person.get_primary_name()] + person.get_alternate_names():
                    sname = find_surname_name(handle, name.serialize())
                    if sname in changelist:
                        change = True
                        for surn in name.get_surname_list():
                            sname = self.name_cap(surn.get_surname())
                            surn.set_surname(sname)
                if change:
                    # cursor.update(handle, person.serialize())
                    self.db.commit_person(person, self.trans)

        self.db.enable_signals()
        self.db.request_rebuild()
        # FIXME: this probably needs to be removed, and bookmarks
        # should always be rebuilt on a commit_person via signals
        # self.parent.bookmarks.redraw()
        self.close()
        self.cb()


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class ChangeNamesOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2005-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2012       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python Modules
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".Db")

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import URL_WIKISTRING

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
_MINVERSION = 9
_DBVERSION = 14


# -------------------------------------------------------------------------
#
# Importing data into the currently open database.
# Ability to import .grdb files removed in accordance with
# http://gramps.1791082.n4.nabble.com/Status-of-GEPS-023-tp4329141p4329746.html
#
# -------------------------------------------------------------------------
def importData(database, filename, user):
    migrate_dict = {
        "gramps_wiki_migrate_two_to_three_url": URL_WIKISTRING
        + "Gramps_4.2_Wiki_Manual_-_Manage_Family_Trees#"
        "Moving_a_Gramps_2.2_databases_to_Gramps_3.x"
    }
    user.notify_error(
        _("%s could not be opened") % filename,
        _(
            "The Database version is not supported "
            "by this version of Gramps."
            "You should use an old copy of Gramps at "
            "version 3.0.x and import your database into "
            "that version. You should then export a copy "
            "of your data to Gramps XML (Family Tree). "
            "Then you should upgrade to the latest "
            "version of Gramps (for example this version), "
            "create a new empty database and import the "
            "Gramps XML into that version. "
            "Please refer to:"
            "%(gramps_wiki_migrate_two_to_three_url)s"
        )
        % migrate_dict,
    )
    return



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Brian G. Matherly
# Copyright (C) 2013       John Ralls
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
GrampsLocale encapsulates localization and locale
initialization. There's a default locale object but more can be
created with a different locale and translation binding to enable
generating reports in languages and locales other than the one used
for the user interface.
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
import gettext
import sys
import os
import locale
import logging
import time

from binascii import hexlify

from .grampstranslation import GrampsTranslations, GrampsNullTranslations

if sys.platform == "darwin":
    from .maclocale import mac_setup_localization

if sys.platform == "win32":
    from .win32locale import win32_locale_init_from_env, win32_locale_bindtextdomain

from .win32locale import _LOCALE_NAMES

LOG = logging.getLogger("." + __name__)
LOG.propagate = True
HAVE_ICU = False
_ICU_ERR = None
_HDLR = None
# GrampsLocale initialization comes before command-line argument
# passing, so one must set the log level directly. The default is
# logging.WARN. Uncomment the following to change it to logging.DEBUG:
# LOG.setLevel(logging.DEBUG)
try:
    from icu import Locale, Collator, ICUError

    HAVE_ICU = True
except ImportError:
    try:
        from PyICU import Locale, Collator, ICUError

        HAVE_ICU = True
    except ImportError as err:
        ERRSTR = str(err)
        # No logger, save the warning message for later.
        _ICU_ERR = (
            f"ICU not loaded because {ERRSTR}. Localization will be "
            "impaired. Use your package manager to install PyICU"
        )

ICU_LOCALES = None
if HAVE_ICU:
    ICU_LOCALES = Locale.getAvailableLocales()
# locales with right-to-left text
_RTL_LOCALES = ("ar", "he")

# locales with less than 70% currently translated
INCOMPLETE_TRANSLATIONS = ("ar", "bg", "ko", "sq", "zh_HK", "zh_TW")


def _check_gformat():
    """
    Some OS environments do not support the locale.nl_langinfo() method
    of determing month names and other date related information.
    """
    try:
        gformat = locale.nl_langinfo(locale.D_FMT).replace("%y", "%Y")
        # Gramps treats dates with '-' as ISO format, so replace separator
        # on locale dates that use '-' to prevent confict
        gformat = gformat.replace("-", "/")
    except locale.Error:
        # Depending on the locale, the value returned for 20th Feb 2009
        # could be '20/2/2009', '20/02/2009', '20.2.2009', '20.02.2009',
        # '20-2-2009', '20-02-2009', '2009/02/20', '2009.02.20',
        # '2009-02-20', or '09-02-20' so to reduce the possible values to
        # test for, make sure both the month and the day are double digits,
        # preferably greater than 12 for human readablity

        timestr = time.strftime("%x", (2005, 10, 25, 1, 1, 1, 1, 1, 1))

        # Gramps treats dates with '-' as ISO format, so replace separator
        # on locale dates that use '-' to prevent confict
        timestr = timestr.replace("-", "/")

        time2fmt_map = {
            "25/10/2005": "%d/%m/%Y",
            "10/25/2005": "%m/%d/%Y",
            "2005/10/25": "%Y/%m/%d",
            "25.10.2005": "%d.%m.%Y",
            "10.25.2005": "%m.%d.%Y",
            "2005.10.25": "%Y.%m.%d",
        }

        try:
            gformat = time2fmt_map[timestr]
        except KeyError:
            gformat = "%d/%m/%Y"  # default value
    return gformat


# ------------------------------------------------------------------------
#
# GrampsLocale Class
#
# ------------------------------------------------------------------------
class GrampsLocale:
    """
    Encapsulate a locale.  This class is a sort-of-singleton: The
    first instance created will query the environment and OSX defaults
    for missing parameters (precedence is parameters passed to the
    constructor, environment variables LANG, LC_COLLATE, LC_TIME,
    etc., and LANGUAGE, OSX defaults settings when that's the
    platform).  Subsequent calls to the constructor with no or
    identical parameters will return the same Grampslocale
    object. Construction with different parameters will result in a
    new GrampsLocale instance with the specified parameters, but any
    parameters left out will be filled in from the first instance.

    :param localedir: The full path to the top level directory containing the
                      translation files. Defaults to sys.prefix/share/locale.

    :param lang: A single locale value which is used for unset locale.LC_FOO
                 settings.

    :param domain: The name of the applicable translation file. The default is
                   "gramps", indicating files in LC_MESSAGES named gramps.mo.

    :param languages: String with a ':'-separated list of two or five character
                      codes corresponding to subidrectries in the localedir,
                      e.g.: "fr" or "zh_CN".
    """

    DEFAULT_TRANSLATION_STR = "default"
    __first_instance = None
    encoding = None

    def __new__(cls, localedir=None, lang=None, domain=None, languages=None):
        if not GrampsLocale.__first_instance:
            cls.__first_instance = super(GrampsLocale, cls).__new__(cls)
            cls.__first_instance.initialized = False
            return cls.__first_instance

        if not cls.__first_instance.initialized:
            raise RuntimeError(
                "Second GrampsLocale created before " "first one was initialized"
            )
        if cls.__first_instance._matches(localedir, lang, domain, languages):
            return cls.__first_instance

        return super(GrampsLocale, cls).__new__(cls)

    def _matches(self, localedir, lang, domain, languages):
        matches = localedir is None or self.localedir == localedir
        matches = matches and (lang is None or self.lang == lang)
        matches = matches and (domain is None or self.localedomain == domain)
        return matches and (
            languages is None or len(languages) == 0 or self.language == languages
        )

    def _init_from_environment(self):
        def _check_locale(loc):
            if not loc[0]:
                return False
            lang = self.check_available_translations(loc[0])
            if not lang and loc[0].startswith("en"):
                loc = ("en_GB", "UTF-8")
                lang = "en_GB"
            if not lang:
                return False
            self.lang = loc[0]
            self.encoding = loc[1]
            self.language = [lang]
            return True

        _failure = False
        try:
            locale.setlocale(locale.LC_ALL, "")
            if not _check_locale(locale.getlocale(category=locale.LC_MESSAGES)):
                LOG.debug("Usable locale not found, localization " "settings ignored.")
                self.lang = "C"
                self.encoding = "ascii"
                self.language = ["en"]
                _failure = True

        except locale.Error as loc_err:
            LOG.debug("Locale error %s, localization settings ignored.", loc_err)
            self.lang = "C"
            self.encoding = "ascii"
            self.language = ["en"]
            _failure = True

        # LC_MESSAGES
        loc_tuple = locale.getlocale(locale.LC_MESSAGES)
        loc = loc_tuple[0]
        if loc:
            language = self.check_available_translations(loc)
            if language:
                self.language = [language]
            else:
                LOG.debug("No translation for LC_MESSAGES locale %s", loc)

        if HAVE_ICU:
            self.calendar = locale.getlocale(locale.LC_TIME)[0] or self.lang[:5]
            self.collation = locale.getlocale(locale.LC_COLLATE)[0] or self.lang[:5]
        else:
            loc = locale.getlocale(locale.LC_TIME)
            if loc and self.check_available_translations(loc[0]):
                self.calendar = ".".join(loc)
            else:
                self.calendar = self.lang

        loc = locale.getlocale(locale.LC_COLLATE)
        if loc and loc[0]:
            self.collation = ".".join(loc)
        else:
            self.collation = self.lang

        if HAVE_ICU and "COLLATION" in os.environ:
            self.collation = os.environ["COLLATION"]

        # $LANGUAGE overrides $LANG, $LC_MESSAGES
        if "LANGUAGE" in os.environ:
            language = [
                x
                for x in [
                    self.check_available_translations(l)
                    for l in os.environ["LANGUAGE"].split(":")
                ]
                if x
            ]
            if language:
                self.language = language
                if not self.lang.startswith(self.language[0]):
                    LOG.debug(
                        "Overiding locale setting '%s' with " "LANGUAGE setting '%s'",
                        self.lang,
                        self.language[0],
                    )
                    self.lang = self.calendar = self.language[0]
            elif _failure:
                LOG.warning("No valid locale settings found, using US English")

        if __debug__:
            LOG.debug(
                "The locale tformat for '%s' is '%s'", self.lang, _check_gformat()
            )

    def __init_first_instance(self):
        """
        Initialize the primary locale from whatever might be
        available. We only do this once, and the resulting
        GrampsLocale is returned by default.
        """
        global _HDLR
        _HDLR = logging.StreamHandler()
        _HDLR.setFormatter(
            logging.Formatter(fmt="%(name)s.%(levelname)s: " "%(message)s")
        )
        LOG.addHandler(_HDLR)

        # Now that we have a logger set up we can issue the icu error if needed.
        if not HAVE_ICU:
            LOG.warning(_ICU_ERR)

        # Even the first instance can be overridden by passing lang
        # and languages to the constructor. If it isn't (which is the
        # expected behavior), do platform-specific setup:
        if not (self.lang and self.language):
            if sys.platform == "darwin":
                mac_setup_localization(self)
            elif sys.platform == "win32":
                win32_locale_init_from_env(self, HAVE_ICU, ICU_LOCALES)
            else:
                self._init_from_environment()
        else:
            self.calendar = self.collation = self.lang

        if not self.lang:
            self.lang = "en_US.UTF-8"
        if not self.language:
            self.language.append("en")
        if not self.localedir and not self.lang.startswith("en"):
            LOG.warning(
                "No translations for %s were found, setting "
                "localization to U.S. English",
                self.localedomain,
            )
            self.lang = "en_US.UTF-8"
            self.language = ["en"]

        # Next, we need to know what is the encoding from the native
        # environment. This is used by python standard library funcions which
        # localize their output, e.g. time.strftime(). NB: encoding is a class variable.
        if not self.encoding:
            self.encoding = locale.getpreferredencoding() or sys.getdefaultencoding()
        LOG.debug("Setting encoding to %s", self.encoding)

        # Make sure that self.lang and self.language are reflected
        # back into the environment for Gtk to use when its
        # initialized. If self.lang isn't 'C', make sure that it has a
        # 'UTF-8' suffix, because that's all that GtkBuilder can
        # digest.

        # Gtk+ has an 'en' po, but we don't. This is worked-around for
        # our GrampsTranslation class but that isn't used to retrieve
        # translations in GtkBuilder (glade), a direct call to libintl
        # (gettext) is. If 'en' is in the translation list it gets
        # skipped in favor of the next language, which can cause
        # inappropriate translations of strings in glade/ui files. To
        # prevent this, if 'en' is in self.language it's the last
        # entry:

        if "en" in self.language:
            self.language = self.language[: self.language.index("en") + 1]

        # Linux note: You'll get unsupported locale errors from Gtk
        # and untranslated strings if the requisite UTF-8 locale isn't
        # installed. This is particularly a problem on Debian and
        # Debian-derived distributions which by default don't install
        # a lot of locales.
        lang = locale.normalize(self.language[0] if self.language[0] else "C")
        check_lang = lang.split(".")
        if not check_lang[0] in ("C", "en"):
            if len(check_lang) < 2 or check_lang[1] not in ("utf-8", "UTF-8"):
                lang = ".".join((check_lang[0], "UTF-8"))

        os.environ["LANG"] = lang
        # We need to convert 'en' and 'en_US' to 'C' to avoid confusing
        # GtkBuilder when it's retrieving strings from our Glade files
        # since we have neither an en.po nor an en_US.po.

        os.environ["LANGUAGE"] = ":".join(self.language)

        # GtkBuilder uses GLib's g_dgettext wrapper, which oddly is bound
        # with locale instead of gettext. Win32 doesn't support bindtextdomain.
        if self.localedir:
            if sys.platform == "win32":
                win32_locale_bindtextdomain(
                    self.localedomain.encode("utf-8"), self.localedir.encode("utf-8")
                )
            else:
                # bug12278, _build_popup_ui() under linux and macOS
                locale.textdomain(self.localedomain)
                locale.bindtextdomain(self.localedomain, self.localedir)

        self.rtl_locale = False
        if self.language[0] in _RTL_LOCALES:
            self.rtl_locale = True  # right-to-left

    def _init_secondary_locale(self):
        """
        Init a secondary locale. Secondary locales are used to provide
        an alternate localization to the one used for the UI; for
        example, some reports offer the option to use a different
        language.

        This GrampsLocale class does no caching of the secondary locale.
        If any caching is desired it must be done externally.
        """
        if not self.localedir:
            LOG.warning("No Localedir provided, unable to find translations")

        _first = self._GrampsLocale__first_instance
        if not self.localedomain:
            if _first.localedomain:
                self.localedomain = _first.localedomain
            else:
                self.localedomain = "gramps"

        if not self.lang and _first.lang:
            self.lang = _first.lang

        if not self.language:
            if self.lang:
                trans = self.check_available_translations(self.lang)
            if trans:
                self.language = [trans]

        if not self.language and _first.language:
            self.language = _first.language

        self.calendar = self.collation = self.lang

        self.rtl_locale = False
        if self.language[0] in _RTL_LOCALES:
            self.rtl_locale = True  # right-to-left

    def __init__(self, localedir=None, lang=None, domain=None, languages=None):
        """
        Init a GrampsLocale. Run __init_first_instance() to set up the
        environment if this is the first run. Return __first_instance
        otherwise if called without arguments.
        """
        global _HDLR
        # initialized is special, used only for the "first instance",
        # and created by __new__(). It's used to prevent re-__init__ing
        # __first_instance when __new__() returns its pointer.
        if hasattr(self, "initialized") and self.initialized:
            return
        self.localedir = None
        self.collator = None
        self.collation = None
        self.calendar = None
        self.rtl_locale = False
        _first = self._GrampsLocale__first_instance
        # Everything breaks without localedir, so get that set up
        # first.  Warnings are logged in __init_first_instance or
        # _init_secondary_locale if this comes up empty.
        if localedir and os.path.exists(os.path.abspath(localedir)):
            self.localedir = localedir
        elif (
            _first
            and hasattr(_first, "localedir")
            and _first.localedir
            and os.path.exists(os.path.abspath(_first.localedir))
        ):
            self.localedir = _first.localedir
        else:
            LOG.warning(
                "Missing or invalid localedir %s; no translations"
                " will be available.",
                repr(localedir),
            )
        self.lang = lang
        self.localedomain = domain or "gramps"
        if languages:
            self.language = [
                x
                for x in [
                    self.check_available_translations(l) for l in languages.split(":")
                ]
                if x
            ]
        else:
            self.language = None

        if self == _first:
            self._GrampsLocale__init_first_instance()
        else:
            self._init_secondary_locale()

        self.icu_locales = {}
        if HAVE_ICU:
            self.icu_locales["default"] = Locale.createFromName(self.lang)
            if self.collation and self.collation != self.lang:
                collation = Locale.createFromName(self.collation)
                self.icu_locales["collation"] = collation
            else:
                self.icu_locales["collation"] = self.icu_locales["default"]
            try:
                collation = self.icu_locales["collation"]
                self.collator = Collator.createInstance(collation)
            except ICUError as icu_err:
                LOG.warning("Unable to create collator: %s", str(icu_err))
                self.collator = None

        try:
            self.translation = self._get_translation(
                self.localedomain, self.localedir, self.language
            )
        except ValueError:
            LOG.warning(
                "Unable to find translation for languages in %s, " "using US English",
                ":".join(self.language),
            )
            self.translation = GrampsNullTranslations()
            self.translation.lang = "en"

        if _HDLR:
            LOG.removeHandler(_HDLR)
            _HDLR = None
        self._dd = self._dp = None
        # Guards against running twice on the first instance.
        self.initialized = True

    def _get_translation(self, domain=None, localedir=None, languages=None):
        """
        Get a translation of one of our classes. Doesn't return the
        singleton so that it can be used by get_addon_translation()
        """
        if not domain:
            domain = self.localedomain
        if not languages:
            languages = self.language
        if not localedir:
            localedir = self.localedir

        for lang in languages:
            if gettext.find(domain, localedir, [lang]):
                translator = gettext.translation(
                    domain,
                    localedir=localedir,
                    languages=[lang],
                    class_=GrampsTranslations,
                )
                translator.lang = lang
                return translator

            if lang.startswith("en") or lang.startswith("C"):
                translator = GrampsNullTranslations()
                translator.lang = "en"
                return translator

        if not languages or len(languages) == 0:
            LOG.warning("No language provided, using US English")
        else:
            lang_str = ":".join(languages)
            raise ValueError(f"No usable translations in {lang_str}" f" for {domain}")
        translator = GrampsNullTranslations()
        translator.lang = "en"
        return translator

    def _get_language_string(self, lang_code):
        """
        Given a language code of the form "lang_region", return a text string
        representing that language.
        """
        try:
            lang = _LOCALE_NAMES[lang_code][2]
        except KeyError:
            try:
                lang = _LOCALE_NAMES[lang_code[:2]][2]
            except KeyError:
                LOG.debug("Gramps has no translation for %s", lang_code)
                lang = None
        except IndexError:
            LOG.debug("Bad Index for tuple %s\n", _LOCALE_NAMES[lang_code][0])
            lang = None

        if lang:
            return self.translation.gettext(lang)
        return lang

    # -------------------------------------------------------------------------
    #
    # Properties
    #
    # -------------------------------------------------------------------------
    @property
    def date_displayer(self):
        """
        Return the locale's date displayer; if it hasn't already been
        cached, set it from datehandler.LANG_TO_DISPLAY. If one isn't
        available for the selected locale, attempt to fall back on the
        first_instance's locale before settling on the 'C' displayer.

        .. note:: This is the getter for the date_displayer property
        """
        if self._dd:
            return self._dd
        # PyLint will whine about these not being at the top level but putting
        # it there creates a circular import.
        from ..config import config
        from ..datehandler import LANG_TO_DISPLAY as displayers

        try:
            val = config.get("preferences.date-format")
        except AttributeError:
            val = 0

        _first = self._GrampsLocale__first_instance
        if self.calendar in displayers:
            self._dd = displayers[self.calendar](val)
        elif self.calendar and self.calendar[:2] in displayers:
            self._dd = displayers[self.calendar[:2]](val)
        elif self != _first and _first.calendar in displayers:
            self._dd = displayers[_first.calendar](val, blocale=self)
        elif self != _first and _first.calendar and _first.calendar[:2] in displayers:
            self._dd = displayers[_first.calendar[:2]](val, blocale=self)
        else:
            self._dd = displayers["C"](val, blocale=self)

        return self._dd

    @property
    def date_parser(self):
        """
        Return the locale's date parser; if it hasn't already been
        cached, set it from datehandler.LANG_TO_PARSER. If one isn't
        available for the selected locale, attempt to fall back on the
        first_instance's locale before settling on the 'C' parser.

        .. note:: This is the getter for the date_parser property
        """
        if self._dp:
            return self._dp
        # PyLint will whine about this not being at the top level but putting
        # it there creates a circular import.
        from ..datehandler import LANG_TO_PARSER as parsers

        _first = self._GrampsLocale__first_instance
        if self.calendar in parsers:
            self._dp = parsers[self.calendar]()
        elif self.calendar and self.calendar[:2] in parsers:
            self._dp = parsers[self.calendar]()
        elif self != _first and _first.calendar in parsers:
            self._dp = parsers[_first.calendar]()
        elif self != _first and _first.calendar and _first.calendar[:2] in parsers:
            self._dp = parsers[_first.calendar[:2]]()
        else:
            self._dp = parsers["C"]()

        return self._dp

    # -------------------------------------------------------------------------
    #
    # Public Functions
    #
    # -------------------------------------------------------------------------

    def get_localedomain(self):
        """
        Get the LOCALEDOMAIN used for the Gramps application.
        Required by gui/glade.py to pass to Gtk.Builder
        """
        return self.localedomain

    def get_language_list(self):
        """
        Return the list of configured languages.  Used by
        ViewManager.check_for_updates to select the language for the
        addons descriptions.
        """
        return self.language

    def locale_code(self):
        """
        Return the locale code.
        """
        return self.lang

    def get_addon_translator(self, filename, domain="addon", languages=None):
        """
        Get a translator for an addon.

        :param filename: filename of a file in directory with full path, or
                         None to get from self.
        :param domain: the name of the .mo file under the LANG/LC_MESSAGES dir
        :param languages: a list of languages to force
        :returns: a gettext.translation object

        Example::

        _ = glocale.get_addon_translator(languages=["fr_BE.utf8"]).gettext

        .. seealso:: the python gettext documentation.

        Assumes path/filename = path/locale/LANG/LC_MESSAGES/addon.mo.
        """
        gramps_translator = self._get_translation(languages=languages)

        path = self.localedir
        if filename:
            path = os.path.join(os.path.dirname(os.path.abspath(filename)), "locale")
        if languages:
            addon_translator = self._get_translation(domain, path, languages=languages)
        else:
            addon_translator = self._get_translation(domain, path)
        gramps_translator.add_fallback(addon_translator)
        return gramps_translator  # with a language fallback

    def get_available_translations(self, localedir=None, localedomain=None):
        """
        Get a list of available translations.

        :returns: A list of translation languages.
        :rtype: unicode[]

        """
        languages = ["en"]

        if not localedir and self.localedir:
            localedir = self.localedir
        else:
            return languages

        if not localedomain and self.localedomain:
            localedomain = self.localedomain
        else:
            localedomain = "gramps"

        for langdir in os.listdir(self.localedir):
            mofilename = os.path.join(
                localedir, langdir, "LC_MESSAGES", f"{localedomain}.mo"
            )
            if os.path.exists(mofilename):
                languages.append(langdir)

        languages.sort()

        return languages

    def check_available_translations(self, loc):
        """
        Test a locale for having a translation available
        locale -- string with standard language code, locale code, or name
        """
        if not self.localedir:
            return None
        # Note that this isn't a typo for self.language; self.language
        # is cached so we don't have to query the file system every
        # time this function is called.
        if not hasattr(self, "languages"):
            self.languages = self.get_available_translations()

        if not loc:
            return None

        if loc[:5] in self.languages:
            return loc[:5]
        # US English is the outlier, all other English locales want real English:
        if loc[:2] == "en" and loc[:5] != "en_US":
            return "en_GB"
        if loc[:2] in self.languages:
            return loc[:2]
        return None

    def get_language_dict(self):
        """
        return a dictionary of language names : codes for use by language
        pickers.
        """
        return {
            self._get_language_string(code): code
            for code in self.get_available_translations()
            if self._get_language_string(code)
        }

    def trans_objclass(self, objclass_str):
        """
        Translates objclass_str into "... %s", where objclass_str
        is 'Person', 'person', 'Family', 'family', etc.
        """
        _ = self.translation.gettext
        objclass = objclass_str.lower()
        if objclass == "person":
            retval = _("the person")
        elif objclass == "family":
            retval = _("the family")
        elif objclass == "place":
            retval = _("the place")
        elif objclass == "event":
            retval = _("the event")
        elif objclass == "repository":
            retval = _("the repository")
        elif objclass == "note":
            retval = _("the note")
        elif objclass == "media":
            retval = _("the media")
        elif objclass == "source":
            retval = _("the source")
        elif objclass == "filter":
            retval = _("the filter")
        elif objclass == "citation":
            retval = _("the citation")
        else:
            retval = _("See details")

        return retval

    def sort_key(self, string):
        """
        Return a value suitable to pass to the "key" parameter of sorted()
        """

        if HAVE_ICU and self.collator:
            # ICU can digest strings and unicode
            # Use hexlify() as to make a consistent string, fixing bug #10077
            key = self.collator.getCollationKey(string)
            return hexlify(key.getByteArray()).decode()

        if isinstance(string, bytes):
            string = string.decode("utf-8", "replace")
        try:
            key = locale.strxfrm(string)
        except ICUError as icu_err:
            LOG.warning(
                "Failed to obtain key for %s because %s", self.collation, str(icu_err)
            )
            return string
        return key

    def get_collation(self):
        """
        Return the collation without any character encoding.
        """
        if self.collation is not None:
            return self.collation.split(".")[0]
        return "C"

    def strcoll(self, string1, string2):
        """
        Given two localized strings, compare them and return -1 if
        string1 would sort first, 1 if string2 would, and 0 if
        they are the same.
        """
        key1 = self.sort_key(string1)
        key2 = self.sort_key(string2)
        return -1 if key1 < key2 else (1 if key1 > key2 else 0)

    def get_date(self, date):
        """
        Return a string representing the date appropriate for the language being
        translated.

        :param date: The date to be represented.
        :type date: :class:`~gen.lib.date.Date`
        :returns: The date as text in the proper language.
        :rtype: unicode
        """
        return self.date_displayer.display(date)

    def get_type(self, name):
        """
        Return a string representing the name appropriate for the language being
        translated.

        :param name: The name type to be represented.
        :returns: The name as text in the proper language.
        :rtype: unicode
        """
        # PyLint will complain about this not being at top level but putting it there
        # creates a circular import.
        from ..lib.grampstype import GrampsType

        return GrampsType.xml_str(name)

    def format_string(self, fmt, val, grouping=False, monetary=False):
        """
        Format a string in the current numeric locale. See python's
        locale.format_string for details.  ICU's message formatting codes are
        incompatible with locale's, so just use locale.format_string
        for now.
        """
        return locale.format_string(fmt, val, grouping, monetary)

    def float(self, val):
        """
        Parse a string to a floating point number. Uses locale.atof(),
        in future with ICU present will use icu.NumberFormat.parse().
        """
        try:
            return locale.atof(val)
        except ValueError:
            point = locale.localeconv()["decimal_point"]
            sep = locale.localeconv()["thousands_sep"]
            try:
                if point == ",":
                    return locale.atof(val.replace(" ", sep).replace(".", sep))
                if point == ".":
                    return locale.atof(val.replace(" ", sep).replace(",", sep))
                return None
            except ValueError:
                return None



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Brian G. Matherly
# Copyright (C) 2008  Jerome Rapinat
# Copyright (C) 2008  Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gen.filters.rules/_HasLDSBase.py

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from . import Rule


# -------------------------------------------------------------------------
#
# HasLDSBase
#
# -------------------------------------------------------------------------
class HasLDSBase(Rule):
    """Rule that checks for object with a LDS event"""

    labels = [_("Number of instances:"), _("Number must be:")]
    name = "Objects with LDS events"
    description = "Matches objects with LDS events"
    category = _("General filters")

    def prepare(self, db, user):
        # things we want to do just once, not for every handle
        if self.list[1] == "less than":
            self.count_type = 0
        elif self.list[1] == "greater than":
            self.count_type = 2
        else:
            self.count_type = 1  # "equal to"

        self.userSelectedCount = int(self.list[0])

    def apply(self, db, obj):
        count = len(obj.get_lds_ord_list())
        if self.count_type == 0:  # "less than"
            return count < self.userSelectedCount
        elif self.count_type == 2:  # "greater than"
            return count > self.userSelectedCount
        # "equal to"
        return count == self.userSelectedCount



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    AddressBookListPage
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import FULLCLEAR

LOG = logging.getLogger(".NarrativeWeb")
_ = glocale.translation.sgettext
getcontext().prec = 8


class AddressBookListPage(BasePage):
    """
    Create the index for addresses.
    """

    def __init__(self, report, the_lang, the_title, has_url_addr_res):
        """
        @param: report           -- The instance of the main report class
                                    for this report
        @param: the_lang         -- The lang to process
        @param: the_title        -- The title page related to the language
        @param: has_url_addr_res -- The url, address and residence to use
                                    for the report
        """
        BasePage.__init__(self, report, the_lang, the_title)

        # Name the file, and create it
        output_file, sio = self.report.create_file("addressbook")

        # Add xml, doctype, meta and stylesheets
        result = self.write_header(_("Address Book"))
        addressbooklistpage, dummy_head, dummy_body, outerwrapper = result

        # begin AddressBookList division
        with Html("div", class_="content", id="AddressBookList") as addressbooklist:
            outerwrapper += addressbooklist

            # Address Book Page message
            msg = self._(
                "This page contains an index of all the individuals "
                "in the database, sorted by their surname, with one "
                "of the following: Address, Residence, or Web Links. "
                "Selecting the person&#8217;s name will take you "
                "to their individual Address Book page."
            )
            addressbooklist += Html("p", msg, id="description")

            # begin Address Book table
            with Html(
                "table", class_="infolist primobjlist addressbook " + self.dir
            ) as table:
                addressbooklist += table

                thead = Html("thead")
                table += thead

                trow = Html("tr")
                thead += trow

                trow.extend(
                    Html("th", label, class_=colclass, inline=True)
                    for (label, colclass) in [
                        ["&nbsp;", "ColumnRowLabel"],
                        [self._("Full Name"), "ColumnName"],
                        [self._("Address"), "ColumnAddress"],
                        [self._("Residence"), "ColumnResidence"],
                        [self._("Web Links"), "ColumnWebLinks"],
                    ]
                )

                tbody = Html("tbody")
                table += tbody

                index = 1
                for (
                    dummy_sort_name,
                    person_handle,
                    has_add,
                    has_res,
                    has_url,
                ) in has_url_addr_res:
                    address = None
                    residence = None
                    weblinks = None

                    # has address but no residence event
                    if has_add and not has_res:
                        address = "X"

                    # has residence, but no addresses
                    elif has_res and not has_add:
                        residence = "X"

                    # has residence and addresses too
                    elif has_add and has_res:
                        address = "X"
                        residence = "X"

                    # has Web Links
                    if has_url:
                        weblinks = "X"

                    trow = Html("tr")
                    tbody += trow

                    trow.extend(
                        Html("td", data or "&nbsp;", class_=colclass, inline=True)
                        for (colclass, data) in [
                            ["ColumnRowLabel", index],
                            ["ColumnName", self.addressbook_link(person_handle)],
                            ["ColumnAddress", address],
                            ["ColumnResidence", residence],
                            ["ColumnWebLinks", weblinks],
                        ]
                    )
                    index += 1

        # Add footer and clearline
        footer = self.write_footer(None)
        outerwrapper += (FULLCLEAR, footer)

        # send the page out for processing
        # and close the file
        self.xhtml_writer(addressbooklistpage, output_file, sio, 0)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2010,2017  Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Media object for Gramps.
"""

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
import os
from urllib.parse import urlparse
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .primaryobj import PrimaryObject
from .citationbase import CitationBase
from .notebase import NoteBase
from .datebase import DateBase
from .attrbase import AttributeBase
from .tagbase import TagBase
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

LOG = logging.getLogger(".citation")


# -------------------------------------------------------------------------
#
# Media class
#
# -------------------------------------------------------------------------
class Media(CitationBase, NoteBase, DateBase, AttributeBase, PrimaryObject):
    """
    Container for information about an image file, including location,
    description and privacy.
    """

    def __init__(self, source=None):
        """
        Initialize a Media.

        If source is not None, then object is initialized from values of the
        source object.

        :param source: Object used to initialize the new object
        :type source: Media
        """
        PrimaryObject.__init__(self, source)
        CitationBase.__init__(self, source)
        NoteBase.__init__(self, source)
        DateBase.__init__(self, source)
        AttributeBase.__init__(self, source)

        if source:
            self.path = source.path
            self.mime = source.mime
            self.desc = source.desc
            self.checksum = source.checksum
            self.thumb = source.thumb
        else:
            self.path = ""
            self.mime = ""
            self.desc = ""
            self.checksum = ""
            self.thumb = None

    def serialize(self, no_text_date=False):
        """
        Convert the data held in the event to a Python tuple that
        represents all the data elements.

        This method is used to convert the object into a form that can easily
        be saved to a database.

        These elements may be primitive Python types (string, integers),
        complex Python types (lists or tuples, or Python objects. If the
        target database cannot handle complex types (such as objects or
        lists), the database is responsible for converting the data into
        a form that it can use.

        :returns: Returns a python tuple containing the data that should
                  be considered persistent.
        :rtype: tuple
        """
        return (
            self.handle,
            self.gramps_id,
            self.path,
            self.mime,
            self.desc,
            self.checksum,
            AttributeBase.serialize(self),
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            self.change,
            DateBase.serialize(self, no_text_date),
            TagBase.serialize(self),
            self.private,
        )

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        from .attribute import Attribute
        from .date import Date

        return {
            "type": "object",
            "title": _("Media"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "handle": {"type": "string", "maxLength": 50, "title": _("Handle")},
                "gramps_id": {"type": "string", "title": _("Gramps ID")},
                "path": {"type": "string", "title": _("Path")},
                "mime": {"type": "string", "title": _("MIME")},
                "desc": {"type": "string", "title": _("Description")},
                "checksum": {"type": "string", "title": _("Checksum")},
                "attribute_list": {
                    "type": "array",
                    "items": Attribute.get_schema(),
                    "title": _("Attributes"),
                },
                "citation_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Citations"),
                },
                "note_list": {
                    "type": "array",
                    "items": {"type": "string"},
                    "title": _("Notes"),
                },
                "change": {"type": "integer", "title": _("Last changed")},
                "date": {
                    "oneOf": [{"type": "null"}, Date.get_schema()],
                    "title": _("Date"),
                },
                "tag_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Tags"),
                },
                "private": {"type": "boolean", "title": _("Private")},
            },
        }

    def unserialize(self, data):
        """
        Convert the data held in a tuple created by the serialize method
        back into the data in a Media structure.

        :param data: tuple containing the persistent data associated the object
        :type data: tuple
        """
        (
            self.handle,
            self.gramps_id,
            self.path,
            self.mime,
            self.desc,
            self.checksum,
            attribute_list,
            citation_list,
            note_list,
            self.change,
            date,
            tag_list,
            self.private,
        ) = data

        AttributeBase.unserialize(self, attribute_list)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        DateBase.unserialize(self, date)
        TagBase.unserialize(self, tag_list)
        return self

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.path, self.mime, self.desc, self.gramps_id]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return self.attribute_list

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer to citations.

        :returns: Returns the list of child secondary child objects that may
                  refer to citations.
        :rtype: list
        """
        return self.attribute_list

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return self.attribute_list

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        return (
            self.get_referenced_note_handles()
            + self.get_referenced_tag_handles()
            + self.get_referenced_citation_handles()
        )

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return self.get_citation_child_list()

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this media object.

        Lost: handle, id, file, date of acquisition.

        :param acquisition: The media object to merge with the present object.
        :type acquisition: Media
        """
        self._merge_privacy(acquisition)
        self._merge_attribute_list(acquisition)
        self._merge_note_list(acquisition)
        self._merge_citation_list(acquisition)
        self._merge_tag_list(acquisition)

    def set_mime_type(self, mime_type):
        """
        Set the MIME type associated with the Media.

        :param mime_type: MIME type to be assigned to the object
        :type mime_type: str
        """
        self.mime = mime_type

    def get_mime_type(self):
        """
        Return the MIME type associated with the Media.

        :returns: Returns the associated MIME type
        :rtype: str
        """
        return self.mime

    def set_path(self, path):
        """Set the file path to the passed path."""
        res = urlparse(path)
        if res.scheme == "" or res.scheme == "file":
            self.path = os.path.normpath(path)
        else:
            # The principal case this path caters for is where the scheme is
            # 'http' or 'https'. It would be possible to do some more extensive
            # checking or processing, but for now we just store the reference
            self.path = path

    def get_path(self):
        """Return the file path."""
        return self.path

    def set_description(self, text):
        """Set the description of the image."""
        self.desc = text

    def get_description(self):
        """Return the description of the image."""
        return self.desc

    def set_checksum(self, text):
        """Set the checksum of the image."""
        self.checksum = text

    def get_checksum(self):
        """Return the checksum of the image."""
        return self.checksum
